# -*- coding: utf-8 -*-
# @Author: Pengyao Ping
# @Date:   2023-02-16 11:01:06
# @Last Modified by:   Pengyao Ping
# @Last Modified time: 2023-05-19 09:59:43

import os
from Bio import SeqIO
import pandas as pd
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from noise2read.isolates_correction import IsolatesErrorCorrection
from noise2read.classifier import MLClassifier 
import copy
from noise2read.utils import *
from noise2read.reads2vectors import Reads2Vectors
# import shutil

class ErrorCorrection():
    """
    A class contains error correction related functions
    """
    def __init__(self, logger, config):
        """
        initialize the ErrorCorrection class

        Args:
            logger (class): customized logging
            config (class): parameters below setting using configparser
        """
        self.logger = logger
        self.config = config
        file_type = parse_file_type(config.input_file)
        if ".gz" in file_type:
            self.out_file_tye = file_type.split(".gz")[0]
        else:
            self.out_file_tye = file_type
        bases = config.input_file.split('/')[-1]
        self.base = bases.split('.')

    def all_in_one_correction(self, isolates_file, non_isolates_file, unique_seqs, genuine_df, negative_df, ambiguous_df, high_ambiguous_df):
        """
        the main function intergrates all_in_one_ambiguous_err_prediction, all_in_one_high_ambiguous_err_correction and IsolatesErrorCorrection

        Args:
            isolates_file (str):  The isolates' filename with path.
            non_isolates_file (str):  The non-isolates' filename with path
            unique_seqs (set): a set contains unique reads
            genuine_df (DataFrame): pandas dataframe containing positive samples for training
            negative_df (DataFrame): pandas dataframe containing negative samples for training
            ambiguous_df (DataFrame): pandas dataframe containing ambiguous samples for prediction
            high_ambiguous_df (DataFrame): pandas dataframe containing high ambiguous samples for prediction

        Returns:
            MultiVariables: 
                corrected_file (str): The corrected data filename including path.
                tmp_correct: The filename including path of corrected data without prediction high ambiguous errors
                new_negative_df: pandas dataframe containing negative samples from the prediction result of ambiguous errors
        """
        corrected_file = self.config.result_dir + self.base[0] + '_corrected.' + self.out_file_tye  
        if isinstance(high_ambiguous_df, pd.DataFrame):
            genuine_ambi_errs_df, new_negative_df, high_ambi_df = self.all_in_one_ambiguous_err_prediction(unique_seqs, genuine_df, negative_df, ambiguous_df, high_ambiguous_df, edit_dis=1)
            # correct errors
            genuine_corrected_file = self.correct_errors(non_isolates_file, genuine_ambi_errs_df)
            self.logger.info("Genuine and ambiguous errors corrected.")
            
            del genuine_ambi_errs_df, genuine_df, negative_df, ambiguous_df, unique_seqs
            ###############################################################################################
            # IEC = IsolatesErrorCorrection(self.logger, self.config.num_workers, isolates_file, genuine_corrected_file, self.config.result_dir, self.config.iso_change_detail, self.config.min_iters)
            # tmp_corrected_isolates = IEC.bcool_correct_isolates() 
            # tmp_correct = self.config.result_dir + self.base[0] + '_correct_no_high.' + self.out_file_tye  
            # if tmp_corrected_isolates and genuine_corrected_file:
            #     os.system("cat %s %s > %s" % (tmp_corrected_isolates, genuine_corrected_file, tmp_correct))
            # # if os.path.exists(tmp_corrected_isolates):     
            # #     os.system("rm %s" % tmp_corrected_isolates)
            # del IEC
            #####################################################################
            non_isolates_correct = self.all_in_one_high_ambiguous_err_correction(genuine_corrected_file, high_ambi_df)
            del high_ambiguous_df
            self.logger.info("High ambiguous errors corrected.")
            self.logger.info('1nt-edit-distance based Errors Correction finished.')
            self.logger.info("#############################################")
        else:
            genuine_ambi_errs_df, new_negative_df = self.all_in_one_ambiguous_err_prediction(unique_seqs, genuine_df, negative_df, ambiguous_df, high_ambiguous_df=None, edit_dis=1)
            # correct errors
            self.logger.info("Correcting 1nt-edit-distance based Errors")

            non_isolates_correct = self.correct_errors(non_isolates_file, genuine_ambi_errs_df)
            self.logger.info('1nt-edit-distance based Errors Correction Finished')
            del genuine_ambi_errs_df, negative_df, ambiguous_df, unique_seqs
        # # bcool correction
        IEC = IsolatesErrorCorrection(self.logger, self.config.num_workers, isolates_file, non_isolates_correct, self.config.result_dir, self.config.iso_change_detail, self.config.min_iters)
        corrected_isolates = IEC.bcool_correct_isolates() 
        if corrected_isolates and non_isolates_correct:
            os.system("cat %s %s > %s" % (corrected_isolates, non_isolates_correct, corrected_file))
        else:
            self.logger.error("No corrected_isolates and/or non_isolates_correct")
        del IEC
        if os.path.exists(isolates_file):
            os.system("rm %s" % isolates_file)
        if os.path.exists(non_isolates_file):
            os.system("rm %s" % non_isolates_file)
        if os.path.exists(corrected_isolates):   
            os.system("rm %s" % corrected_isolates)
        if os.path.exists(non_isolates_correct):     
            os.system("rm %s" % non_isolates_correct)
        # if os.path.exists(self.config.result_dir + "bcool/"):
        #     shutil.rmtree(self.config.result_dir + "bcool/") 
        # if isinstance(high_ambiguous_df, pd.DataFrame):
        #     del high_ambiguous_df
        #     return corrected_file, tmp_correct, new_negative_df
        # else:
        return corrected_file, new_negative_df

    def all_in_one_ambiguous_err_prediction(self, total_reads, genuine_df, negative_df, ambiguous_df, high_ambiguous_df, edit_dis):
        """
        ambiguous or high ambiguous errors prediction

        Args:
            total_reads (set): set contains all the reads
            genuine_df (DataFrame): pandas dataframe containing positive samples for training
            negative_df (DataFrame): pandas dataframe containing negative samples for training
            ambiguous_df (DataFrame): pandas dataframe containing ambiguous samples for prediction
            high_ambiguous_df (DataFrame): pandas dataframe containing high ambiguous samples for prediction
            edit_dis (int): edit distance 1 or 2

        Returns:
            MultiVariables: genuine_df, new_negative_df, high_ambiguous_df
        """
        RV = Reads2Vectors(self.logger, self.config, edit_dis)
        # RV = Reads2Vectors(self.logger, self.config.num_workers, self.config.result_dir, self.config.read_max_len, self.config.entropy_kmer, self.config.entropy_q, self.config.kmer_freq, self.config.read_type, edit_dis)
        if edit_dis == 1:
            study_name = 'ambiguous_1nt'
        elif edit_dis == 2:
            study_name = 'ambiguous_2nt'

        train_data, train_labels, ambiguous_data = RV.all_in_one_embedding(total_reads, genuine_df, negative_df, ambiguous_df, high_flag=False)  
        del total_reads
        TM = MLClassifier(self.logger, self.config, study_name, train_data, train_labels, ambiguous_data)
        predictions = TM.tunning(self.config.n_trials)
        ambiguous_df.insert(ambiguous_df.shape[1], 'predictand', predictions)

        # merge genuine and ambiguous errors
        grouped_ambiguous_df = ambiguous_df.groupby("idx")
        new_negative_df = pd.DataFrame()
        for name, group_df in grouped_ambiguous_df:
            self.logger.debug(group_df['predictand'])
            # if "N-X" or "X-N" in group_df['ErrorTye'].tolist():
            if len(group_df) > 0:
                pred_index = group_df['predictand'].idxmax()
                new_df = group_df.loc[:, ~group_df.columns.isin(['idx', 'predictand'])]
                entry = copy.deepcopy(new_df.loc[[pred_index]])
                bad_df = group_df.index.isin([pred_index])
                cur_negative = group_df[~bad_df]
                genuine_df = pd.concat([genuine_df, entry], ignore_index=True)
                new_negative_df = pd.concat([new_negative_df, cur_negative], ignore_index=True)
                
        if edit_dis == 1 and self.config.verbose:
            ambiguous_df.to_csv(self.config.result_dir + 'ambiguous_1nt_prediction.csv', index=False)
        elif edit_dis == 2 and self.config.verbose:
            ambiguous_df.to_csv(self.config.result_dir + 'ambiguous_2nt_prediction.csv', index=False)
        del TM, ambiguous_data, ambiguous_df, grouped_ambiguous_df
        #######################################################################################################
        if isinstance(high_ambiguous_df, pd.DataFrame):
            high_train_data, high_train_labels, high_ambiguous_data = RV.high_all_in_one_embedding(genuine_df, negative_df, new_negative_df, high_ambiguous_df)
            study_name = 'high_ambiguous_1nt'
            TM = MLClassifier(self.logger, self.config, study_name, high_train_data, high_train_labels, high_ambiguous_data)
            high_predictions = TM.tunning(self.config.n_trials)
            high_ambiguous_df.insert(high_ambiguous_df.shape[1], 'predictand', high_predictions)
            del RV, TM, high_train_data, high_train_labels, high_ambiguous_data, negative_df
            return genuine_df, new_negative_df, high_ambiguous_df           
        else:
            del negative_df
            return genuine_df, new_negative_df   

    def correct_errors(self, orginal_file, df_data):
        """
        correct errors in the raw to yield corrected dataset

        Args:
            orginal_file (str): raw data filename including path
            df_data (DataFrame): pandas dataframe save predicted result

        Returns:
            str: corrected data filename including path
        """
        record_dict, ori_file_type = parse_data_dict(orginal_file)
        seq2id = {}
        total_name_lst = []
        for id in record_dict:
            seq = str(record_dict[id].seq)
            seq2id.setdefault(seq, []).append(id)
            total_name_lst.append(id)
        err2cor_records = []
        err_name_lst = []
        for row_index, row in df_data.iterrows():
            err_count = row['EndReadCount']
            err_read = row['EndRead']
            cor_read = row['StartRead']
            err_pos = row['ErrorPosition']
            err_tye = row['ErrorTye']
            names = seq2id[err_read]
            endErrorTye = self.error_class(err_tye)
            err_name_lst.extend(names)
            if err_count == len(names):
                for name in names:
                    if ori_file_type == 'fastq' or ori_file_type == 'fq' or ori_file_type == 'fastq.gz' or ori_file_type == 'fq.gz':
                        qual = {}
                        val = record_dict[name].letter_annotations['phred_quality']
                        val.sort()
                        position_val = val[int(len(val)/2)]
                        if endErrorTye == '0':
                            val[err_pos] = position_val
                        elif endErrorTye == '2':
                            del val[err_pos]
                        elif endErrorTye == '1':
                            val.insert(err_pos, position_val)
                        qual['phred_quality'] = val
                        tmp_rec = SeqRecord(Seq(cor_read), id=name, description=record_dict[name].description, letter_annotations=qual)  
                    else:
                        tmp_rec = SeqRecord(Seq(cor_read), id=name, description=record_dict[name].description)  
                    err2cor_records.append(tmp_rec)

        bases = orginal_file.split('/')[-1]
        prefix = bases.split('.')
        err_free_name_lst = list(set(total_name_lst) - set(err_name_lst))
        corrected_err_file = self.config.result_dir + prefix[0] + '_err2cor.' + self.out_file_tye
        with open(corrected_err_file, "w") as handle:
            SeqIO.write(err2cor_records, handle, self.out_file_tye)

        err_free_reads_file = self.config.result_dir + prefix[0] + '_errfree.' + self.out_file_tye
        extract_records(self.config.result_dir, err_free_name_lst, orginal_file, err_free_reads_file)
        
        corrected_file = self.config.result_dir + prefix[0] + '_corrected.' + self.out_file_tye
        os.system("cat %s %s > %s" % (corrected_err_file, err_free_reads_file, corrected_file))
        if os.path.exists(corrected_err_file):
            os.system("rm %s" % corrected_err_file)
        if os.path.exists(err_free_reads_file):
            os.system("rm %s" % err_free_reads_file)
        return corrected_file

    def error_class(self, errorType):
        """
        indicate substution or indel error given an error type 

        Args:
            errorType (str): incidate an error type

        Returns:
            str: a flag to indicate substution or indel error
        """
        substitutions = ['A-G', 'A-C', 'A-T','G-A', 'G-C', 'G-T','C-A', 'C-G', 'C-T','T-A', 'T-C', 'T-G', 'A-N', 'T-N','G-N','C-N','N-A','N-T', 'N-C', 'N-G']
        deletions = ['A-X', 'C-X', 'T-X', 'G-X', 'N-X']
        insertions = ['X-A', 'X-G', 'X-C', 'X-T', 'X-N']
        if errorType in substitutions:
            return '0'
        elif errorType in deletions:
            return '1'
        elif errorType in insertions:
            return '2'
        else:
            return '3'

    def all_in_one_high_ambiguous_err_correction(self, orginal_file, ambi_prediction_result):
        """
        high ambiguous error correction

        Args:
            orginal_file (str): raw data filename including path
            ambi_prediction_result (DataFrame): pandas dataframe save predicted result

        Returns:
            str: corrected data filename including path
        """
        grouped_ambiguous_df = ambi_prediction_result.groupby("idx")
        pred_df = pd.DataFrame()
        for name, group_df in grouped_ambiguous_df:
            # self.logger.debug(group_df['predictand'])
            if len(group_df) == 2:
                preds = group_df['predictand'].tolist()
                pred1 = preds[0]
                pred2 = preds[1]
                if abs(pred1-pred2) >= self.config.proba_deviation:
                    pred_index = group_df['predictand'].idxmax()
                    # pred_index = group_df['predictand'].idxmin()
                    # new_df = group_df.loc[:, ~group_df.columns.isin(['idx', 'predictand'])]
                    entry = group_df.loc[[pred_index]]
                    pred_df = pd.concat([pred_df, entry], ignore_index=True)
        if self.config.verbose:
            pred_df.to_csv(self.config.result_dir + 'high_ambiguous_1nt_prediction.csv', index=False)

        if len(pred_df) > 0:
            corrected_file = self.correct_high_ambiguous_errors(orginal_file, pred_df)
            if os.path.exists(orginal_file):     
                os.system("rm %s" % orginal_file)
            return corrected_file
        else:
            return orginal_file

    ######################################################################################################
    def all_in_one_ed2_correction(self, data_set, total_reads, genuine_df, negative_df, ambiguous_df):       
        """
        predicting and correcting ambiguous errors in 2nt-edit-distance-based read graph

        Args:
            data_set (str): raw data filename including path
            total_reads (set): set contains all the reads
            genuine_df (DataFrame): pandas dataframe containing positive samples for training
            negative_df (DataFrame): pandas dataframe containing negative samples for training
            ambiguous_df (DataFrame): pandas dataframe containing ambiguous samples for prediction

        Returns:
            str: corrected data filename including path
        """
        if not genuine_df.empty and not ambiguous_df.empty and not negative_df.empty:
            genuine_ambi_prediction_result, new_negative_df = self.all_in_one_ambiguous_err_prediction(total_reads, genuine_df, negative_df, ambiguous_df, high_ambiguous_df=None, edit_dis=2)
            correct_file = self.all_in_one_2nt_correct_errors(data_set, genuine_ambi_prediction_result)
            os.system("rm %s" % data_set)
            del genuine_ambi_prediction_result, new_negative_df
            self.logger.info("Error Correction finished.")
            return correct_file
        else:
            return data_set

    def all_in_one_2nt_correct_errors(self, orginal_file, df_data):
        """
        correcting ambiguous errors in 2nt-edit-distance-based read graph

        Args:
            orginal_file (str): raw data filename including path
            df_data (DataFrame): pandas dataframe save predicted result

        Returns:
            str: corrected data filename including path
        """
        record_dict, ori_file_type = parse_data_dict(orginal_file)
        seq2id = {}
        total_name_lst = []
        for id in record_dict:
            seq = str(record_dict[id].seq)
            seq2id.setdefault(seq, []).append(id)
            total_name_lst.append(id)
        err2cor_records = []
        err_name_lst = []
        for row_index, row in df_data.iterrows():
            err_count = row['EndReadCount']
            err_read = row['EndRead']
            cor_read = row['StartRead']
            names = seq2id[err_read]
            err_name_lst.extend(names)
            if err_count == len(seq2id[err_read]):
                for name in names:
                    if ori_file_type == 'fastq' or ori_file_type == 'fq' or ori_file_type == 'fastq.gz' or ori_file_type == 'fq.gz':
                        tmp_rec = SeqRecord(Seq(cor_read), id=name, description=record_dict[name].description, letter_annotations=record_dict[name].letter_annotations)  
                    else:
                        tmp_rec = SeqRecord(Seq(cor_read), id=name, description=record_dict[name].description)  
                    err2cor_records.append(tmp_rec)

        bases = orginal_file.split('/')[-1]
        prefix = bases.split('.')
        err_free_name_lst = list(set(total_name_lst) - set(err_name_lst))
        corrected_err_file = self.config.result_dir + prefix[0] + '_err2cor.' + self.out_file_tye
        with open(corrected_err_file, "w") as handle:
            SeqIO.write(err2cor_records, handle, self.out_file_tye)

        err_free_reads_file = self.config.result_dir + prefix[0] + '_errfree.' + self.out_file_tye
        extract_records(self.config.result_dir, err_free_name_lst, orginal_file, err_free_reads_file)
        corrected_file = self.config.result_dir + prefix[0] + '_corrected.' + self.out_file_tye
        os.system("cat %s %s > %s" % (corrected_err_file, err_free_reads_file, corrected_file))
        if os.path.exists(corrected_err_file):
            os.system("rm %s" % corrected_err_file)
        if os.path.exists(err_free_reads_file):
            os.system("rm %s" % err_free_reads_file)
        return corrected_file

    ######################################################################################################
    def umi_correction(self, data, genuine_df):
        """
        umi correction

        Args:
            data (str): raw data filename including path
            genuine_df (DataFrame): pandas dataframe containing genuine errors

        Returns:
            str: corrected data filename including path
        """
        corrected_file = self.config.result_dir + self.base[0] + '_corrected.' + self.out_file_tye
        # errors_df = self.ambiguous_err_prediction(genuine_csv, ambiguous_csv, negative_csv, edit_dis=1)
        # correct errors
        self.logger.info("Correcting 1nt-edit-distance based Errors")
        # genuine_df = pd.read_csv(genuine_csv)
        corrected_file = self.correct_errors(data, genuine_df)
        self.logger.info('1nt-edit-distance based Errors Correction Finished')
        return corrected_file
    
    def correct_high_ambiguous_errors(self, orginal_file, df_data):
        """
        correcting high ambiguous errors from raw dataset to generate the corrected dataset

        Args:
            orginal_file (str): raw data filename including path
            df_data (DataFrame): pandas dataframe save predicted result

        Returns:
            str: corrected data filename including path
        """
        record_dict, ori_file_type = parse_data_dict(orginal_file)
        seq2id = {}
        total_name_lst = []
        for id in record_dict:
            seq = str(record_dict[id].seq)
            seq2id.setdefault(seq, []).append(id)
            total_name_lst.append(id)
            
        err2cor_records = []
        err_name_lst = []
        
        name2flag = {}
        for row_index, row in df_data.iterrows():
            # err_count = row['EndReadCount']
            err_read = row['EndRead']
            # cor_read = row['StartRead']
            names = seq2id[err_read]
            # to set a flag to prevent one id' read to be changed to diffterent true reads multi times
            for name in names:
                name2flag[name] = True

        for row_index, row in df_data.iterrows():
            err_count = row['EndReadCount']
            err_read = row['EndRead']
            cor_read = row['StartRead']
            err_pos = row['ErrorPosition']
            endErrorTye = self.error_class(row['ErrorTye'])
            names = seq2id[err_read]

            if len(names) >= 1:
                # name = names[0]
                name = 'name'
                for i in names:
                    if name2flag[i]:
                        name = copy.deepcopy(i)
                        name2flag[i] = False
                        break
                    else:
                        continue
                self.logger.debug(name)
                if name != 'name':
                    err_name_lst.append(name)
                    # names.remove(name)
                    # seq2id[err_read] = names
                    if ori_file_type == 'fastq' or ori_file_type == 'fq' or ori_file_type == 'fastq.gz' or ori_file_type == 'fq.gz':
                        qual = {}
                        val = record_dict[name].letter_annotations['phred_quality']
                        val.sort()
                        position_val = val[int(len(val)/2)]
                        if endErrorTye == '0':
                            val[err_pos] = position_val
                        elif endErrorTye == '2':
                            del val[err_pos]
                        elif endErrorTye == '1':
                            val.insert(err_pos, position_val)
                        qual['phred_quality'] = val
                        tmp_rec = SeqRecord(Seq(cor_read), id=name, description=record_dict[name].description, letter_annotations=qual) 
                    else:
                        tmp_rec = SeqRecord(Seq(cor_read), id=name, description=record_dict[name].description)  
                    self.logger.debug(f'{name, err_read}')
                    err2cor_records.append(tmp_rec)
                else:
                    self.logger.warning("This high frequency reads hasn't enough reads to be corrected to predicted countparts, and will ingnore this prediction.")
                
        bases = orginal_file.split('/')[-1]
        prefix = bases.split('.')
        err_free_name_lst = list(set(total_name_lst) - set(err_name_lst))
        corrected_err_file = self.config.result_dir + prefix[0] + '_err2cor.' + self.out_file_tye
        self.logger.debug(f'Number of high ambiguous errors has been corrected: {len(err2cor_records)}')
        
        with open(corrected_err_file, "w") as handle:
            SeqIO.write(err2cor_records, handle, self.out_file_tye)

        err_free_reads_file = self.config.result_dir + prefix[0] + '_errfree.' + self.out_file_tye
        extract_records(self.config.result_dir, err_free_name_lst, orginal_file, err_free_reads_file)
        corrected_file = self.config.result_dir + prefix[0] + '_correct.' + self.out_file_tye
        os.system("cat %s %s > %s" % (corrected_err_file, err_free_reads_file, corrected_file))
        if os.path.exists(corrected_err_file):
            os.system("rm %s" % corrected_err_file)
        if os.path.exists(err_free_reads_file):
            os.system("rm %s" % err_free_reads_file)
        return corrected_file

    ######################################################################################################
    def correct_amplicon_err(self, original_file, genuine_df, negative_df, new_negative_df, amplicon_df, predict_proba):
        """
        correcting amplicon errors from dataset to generate the corrected dataset

        Args:
            original_file (str): data filename including path
            genuine_df (DataFrame): pandas dataframe containing positive samples for training
            negative_df (DataFrame): pandas dataframe containing negative samples for training
            new_negative_df (DataFrame): pandas dataframe containing negative samples from ambiguous prediction result
            amplicon_df (DataFrame): pandas dataframe containing amplicon errors samples for prediction
            predict_proba (float): the threshold to indicate whetheru a sample error-prone or error-free

        Returns:
            _type_: _description_
        """
        study_name = 'amplicon_1nt'
        RV = Reads2Vectors(self.logger, self.config, edit_dis=1)
        # RV = Reads2Vectors(self.logger, self.config.num_workers, self.config.result_dir, self.config.read_max_len, self.config.entropy_kmer, self.config.entropy_q, self.config.kmer_freq, self.config.read_type, edit_dis=1)
        # train_data, train_labels, amplicon_data = RV.all_in_one_embedding(total_reads, genuine_df, negative_df, amplicon_df, high_flag=False)  
        train_data, train_labels, amplicon_data = RV.high_all_in_one_embedding(genuine_df, negative_df, new_negative_df, amplicon_df)
        TM = MLClassifier(self.logger, self.config, study_name, train_data, train_labels, amplicon_data)
        predictions = TM.tunning(self.config.n_trials)
        amplicon_df.insert(amplicon_df.shape[1], 'predictand', predictions)
        if self.config.verbose:
            amplicon_df.to_csv(os.path.join(self.config.result_dir, 'amplicon_predition.csv'), index = False)
        grouped_amplicon_df = amplicon_df.groupby("idx")
        pred_df = pd.DataFrame()
        for name, group_df in grouped_amplicon_df:
            # self.logger.debug(group_df)
            if len(group_df) == 1:
                pred = group_df['predictand'].values[0]
                if pred >= predict_proba:
                    pred_df = pd.concat([pred_df, group_df], ignore_index=True)
            elif len(group_df) >= 2:
                # high_index = pred_df = pd.concat([pred_df, group_df], ignore_index=True)['StartReadCount'].idxmax()
                pred_index = group_df['predictand'].idxmax()
                entry = group_df.loc[[pred_index]]
                # self.logger.debug(len(entry))
                pred = entry['predictand'].values[0]
                if pred >= predict_proba:
                    pred_df = pd.concat([pred_df, entry], ignore_index=True)
        # corrected_file = self.config.result_dir + self.base[0] + '_corrected.' + self.out_file_tye 
        del RV, TM, train_data, train_labels, amplicon_data
        if len(pred_df) > 1:
            correct_file = self.correct_errors(original_file, pred_df)
            if os.path.exists(original_file):     
                os.system("rm %s" % original_file)
            return correct_file
        else:
            return original_file

'''
    def correct_amplicon_errors(self, orginal_file, df_data):
        """
        correcting 2nt ambiguous errors from raw dataset to generate the corrected dataset

        Args:
            orginal_file (str): raw data filename including path
            df_data (DataFrame): pandas dataframe save predicted result

        Returns:
            str: corrected data filename including path
        """
        record_dict, ori_file_type = parse_data_dict(orginal_file)
        seq2id = {}
        total_name_lst = []
        for id in record_dict:
            seq = str(record_dict[id].seq)
            seq2id.setdefault(seq, []).append(id)
            total_name_lst.append(id)
        err2cor_records = []
        err_name_lst = []
        for row_index, row in df_data.iterrows():
            err_count = row['EndReadCount']
            err_read = row['EndRead']
            cor_read = row['StartRead']
 
            names = seq2id[err_read]
            err_name_lst.extend(names)
            if err_count == len(seq2id[err_read]):
                for name in names:
                    if ori_file_type == 'fastq' or ori_file_type == 'fq' or ori_file_type == 'fastq.gz' or ori_file_type == 'fq.gz':
                        tmp_rec = SeqRecord(Seq(cor_read), id=name, description=record_dict[name].description, letter_annotations=record_dict[name].letter_annotations)  
                    else:
                        tmp_rec = SeqRecord(Seq(cor_read), id=name, description=record_dict[name].description)  
                    err2cor_records.append(tmp_rec)

        bases = orginal_file.split('/')[-1]
        prefix = bases.split('.')
        err_free_name_lst = list(set(total_name_lst) - set(err_name_lst))
        corrected_err_file = self.config.result_dir + prefix[0] + '_err2cor.' + self.out_file_tye
        with open(corrected_err_file, "w") as handle:
            SeqIO.write(err2cor_records, handle, self.out_file_tye)

        err_free_reads_file = self.config.result_dir + prefix[0] + '_errfree.' + self.out_file_tye
        extract_records(self.config.result_dir, err_free_name_lst, orginal_file, err_free_reads_file)
        corrected_file = self.config.result_dir + prefix[0] + '_correct.' + self.out_file_tye
        os.system("cat %s %s > %s" % (corrected_err_file, err_free_reads_file, corrected_file))
        return corrected_file
'''


# -*- coding: utf-8 -*-
# @Author: Pengyao Ping
# @Date:   2023-02-16 11:01:06
# @Last Modified by:   Pengyao Ping
# @Last Modified time: 2023-02-16 11:11:32

from Bio import SeqIO
import os
import editdistance
import operator
from collections import Counter
from noise2read.utils import *
from pathlib import Path
from Bio.SeqRecord import SeqRecord

class IsolatesErrorCorrection():
    def __init__(self, 
            logger, 
            num_workers,
            isolates, 
            correct_non_isolates,
            output_dir,
            iso_change_detail,
            min_iters):
        self.logger = logger
        self.num_workers = num_workers
        self.isolates = isolates
        self.correct_non_isolates = correct_non_isolates
        self.output_dir = output_dir
        bases = self.isolates.split('/')[-1]
        self.base = bases.split('.')
        self.iso_change_detail = iso_change_detail
        self.min_iters = min_iters
        
    def bcool_correct_isolates(self):
        # # bcool correction
        self.logger.info("Correcting Isolated nodes using bcool.")
        bcool_dir = self.output_dir + "bcool/input/"
        if not Path(bcool_dir).is_dir():
            os.makedirs(bcool_dir)
        iso_file_tye = parse_file_type(self.isolates)
        if iso_file_tye == "fastq":
            read_fa = bcool_dir + "input.fasta"
            # self.covert_fq2fa(self.isolates, read_fa)
            SeqIO.convert(self.isolates, "fastq", read_fa, "fasta-2line")
            # os.system("seqtk seq -a %s > %s" % (self.isolates, read_fa))
            os.system("bcool -u %s -t %s -o %s" % (read_fa, self.num_workers, bcool_dir))
        else:
            os.system("bcool -u %s -t %s -o %s" % (self.isolates, self.num_workers, bcool_dir))
        bcool_isolates = self.output_dir + "bcool/reads_corrected.fa"
        frequency_file = self.output_dir + self.base[0] + '_frequency.txt'

        corrected_isolates = self.select_correction(self.correct_non_isolates, self.isolates, bcool_isolates, self.output_dir + self.base[0], frequency_file)
        # os.system("rm %s" % bcool_isolates)
        self.logger.info("Isolated nodes Correction finished!")
        return corrected_isolates

    def freqency_seq_extraction(self, f_original, f_correct, frequency_file):
        correct_seq2name_dict = {}
        original_name2seq_dict = {}

        ori_record_iterator, _ = parse_data(f_original)
        cor_record_iterator, _ = parse_data(f_correct)
        for ori_rec, cor_rec in zip(ori_record_iterator, cor_record_iterator):
            ori_seq = str(ori_rec.seq)
            cor_seq = str(cor_rec.seq)

            ori_id = str(ori_rec.id)
            cor_id = str(cor_rec.id)

            original_name2seq_dict.setdefault(ori_id, []).append(ori_seq)
            correct_seq2name_dict.setdefault(cor_seq, []).append(cor_id)

        seq_freq = {}
        for seq, name_lst in correct_seq2name_dict.items():
            seq_freq[seq] = len(name_lst)

        seq_freq_order = dict(sorted(seq_freq.items(), key=operator.itemgetter(1), reverse=True))

        with open(frequency_file, 'a') as f:        
            f.write('Top frequency: ')
            f.write('\n')
            f.write(str(list(seq_freq_order.values())))
            f.write('\n')
            for seq, freq in seq_freq_order.items():
                f.write('##########################################')
                f.write('\n')
                f.write(str(freq))
                f.write('\n')
                f.write(seq)
                f.write('\n')
                f.write(str(correct_seq2name_dict[seq]))
                f.write('\n')
                for name in correct_seq2name_dict[seq]:
                    # print(name)
                    f.write(name)
                    f.write('\n')
                    f.write(str(original_name2seq_dict[name]))
                    f.write('\n')
                f.write('##########################################')
        del ori_record_iterator, cor_record_iterator, seq_freq_order, seq_freq, correct_seq2name_dict, original_name2seq_dict
        return

    # def bcool_seqs(self, num_workers, keep_bcool_name_lst, f_bcool, f_original, keep_correct_f):
    def bcool_seqs(self, keep_bcool_name_lst, f_bcool, f_original, keep_correct_f):
        bcool_records_dict, _ = parse_data_dict(f_bcool)
        ori_records_dict, file_type = parse_data_dict(f_original)
        sub_records = []
        for name in keep_bcool_name_lst:
            tmp_rec = SeqRecord(bcool_records_dict[name].seq, id=ori_records_dict[name].id, description=ori_records_dict[name].description, letter_annotations=ori_records_dict[name].letter_annotations)
            sub_records.append(tmp_rec)
        del bcool_records_dict, ori_records_dict
        # shared_objects = copy.deepcopy(bcool_records_dict, ori_records_dict)
        # workerpool and shared_objects use too large memory
        # name_len = len(keep_bcool_name_lst)
        # with WorkerPool(num_workers, shared_objects=shared_objects, start_method='fork') as pool:
        #     with tqdm(total=name_len, desc=self.logger.info("Extract Seqs"), miniters=int(name_len/self.min_iters)) as pbar:   
        #         for tmp_rec in pool.imap(self.bcool_seq, keep_bcool_name_lst):
        #             sub_records.append(tmp_rec)
        #             pbar.update() 
        with open(keep_correct_f, "w") as handle:
            SeqIO.write(sub_records, handle, file_type)  
        # del shared_objects, keep_bcool_name_lst, sub_records
        del keep_bcool_name_lst, sub_records
        return

    def select_correction(self, f_corrected_file, f_original, f_bcool, prefix, frequency_file):
        correct_non_isolates_seqs_lst = []
        correct_record_iterator, f_correct_type = parse_data(f_corrected_file)
        for rec in correct_record_iterator:
            correct_non_isolates_seqs_lst.append(str(rec.seq))

        ori_seqs_lst = []
        cor_seqs_lst = []
        ori_seq2name_dict = {}
        bcool_seq2name_dict = {}
        ori_record_iterator, ori_file_type = parse_data(f_original)
        bcool_record_iterator, bcool_file_type = parse_data(f_bcool)
        for ori_rec, cor_rec in zip(ori_record_iterator, bcool_record_iterator):
            ori_seq = str(ori_rec.seq)
            cor_seq = str(cor_rec.seq)
            ori_name = str(ori_rec.id)
            cor_name = str(cor_rec.id)
            ori_seqs_lst.append(ori_seq)
            cor_seqs_lst.append(cor_seq)
            
            ori_seq2name_dict.setdefault(ori_seq,[]).append(ori_name)
            bcool_seq2name_dict.setdefault(cor_seq,[]).append(cor_name)

        del ori_record_iterator, bcool_record_iterator, correct_record_iterator

        ori_seqs_set = set(ori_seqs_lst)
        cor_seqs_set = set(cor_seqs_lst)

        total_seqs_set = ori_seqs_set | set(correct_non_isolates_seqs_lst)
        del correct_non_isolates_seqs_lst

        self.logger.debug("No. of sequences(isolates) before bcool correction: {}".format(len(ori_seqs_set)))
        if len(ori_seqs_set) == len(ori_seqs_lst):
            self.logger.debug("Each the isolates' frequency equals to one before bcool correction.")
        else:
            self.logger.debug("Before bcool correction, not all the isolates' frequency equals to one.")
        
        self.logger.debug("No. of unique sequences (isolates) after bcool correction: {}".format(len(cor_seqs_set)))

        intersect = ori_seqs_set & cor_seqs_set
        self.logger.debug("No. of the sequences' intersection before and after bcool correction: {}".format(len(intersect)))

        union = cor_seqs_set | ori_seqs_set
        self.logger.debug("No. of the sequences' union for isolates before and after bcool correction: {}".format(len(union)))

        new_sequences = cor_seqs_set - ori_seqs_set
        self.logger.debug("No. of newly generated sequences for isolates before and after bcool correction: {}".format(len(new_sequences)))

        total_intersect = total_seqs_set & cor_seqs_set
        self.logger.debug("No. of the sequences' intersection between corrected isolates and corrected non-isolates and not corrected isolates: {}".format(len(total_intersect)))

        total_union = cor_seqs_set | total_seqs_set
        self.logger.debug("No. of the sequences' union for corrected isolates with corrected non-isolates and not corrected isolates: {}".format(len(total_union)))
        
        select_new_sequences = new_sequences & total_seqs_set
        self.logger.debug("No. of newly generated sequences for corrected isolates existing in corrected non-isolates or not corrected isolates: {}".format(len(select_new_sequences)))

        del total_seqs_set

        cor_seqs_lst_dict = Counter(cor_seqs_lst)
        
        seqs_not_1 = []
        for key, v in cor_seqs_lst_dict.items():
            if v != 1:
                seqs_not_1.append(key)
        seqs_not_1_set = set(seqs_not_1)

        intersect_sequences_set = seqs_not_1_set & total_intersect
        new_sequences_set = seqs_not_1_set & select_new_sequences

        del cor_seqs_lst_dict
        del union
        del new_sequences
        del intersect
        del ori_seqs_lst
        del ori_seqs_set
        del cor_seqs_lst
        del cor_seqs_set
        ############################################################################
        if intersect_sequences_set:
            original_records, _ = parse_data_index(f_original)
            keep_inter_bcool_name_lst = []

            for seq in intersect_sequences_set:
                bcool_names = bcool_seq2name_dict[seq]
                temp_bcool_name_lst = []
                for name in bcool_names:
                    ori_seq = str(original_records[name].seq)
                    if editdistance.eval(seq, ori_seq) <= 1:
                        temp_bcool_name_lst.append(name)                        
                if len(temp_bcool_name_lst) >= 2:
                    keep_inter_bcool_name_lst.extend(temp_bcool_name_lst)
                del temp_bcool_name_lst, bcool_names

            # save these frquency changing reads before and after correction
            len1 = len(keep_inter_bcool_name_lst)
            if len1 and self.iso_change_detail:
                original_bcool_inter_only_fastq_file = prefix + '_original_bcool_inter_only.' + ori_file_type
                extract_records(self.output_dir, keep_inter_bcool_name_lst, f_original, original_bcool_inter_only_fastq_file)

                bcool_inter_only_fastq_file = prefix + '_bcool_inter_only.' + bcool_file_type 
                extract_records(self.output_dir, keep_inter_bcool_name_lst, f_bcool, bcool_inter_only_fastq_file)
                with open(frequency_file, 'w') as f:
                    f.write('===========================================')
                    f.write('\n')
                    f.write('Intersection sequences: ')
                    f.write('\n')
                    f.write('===========================================')
                self.freqency_seq_extraction(original_bcool_inter_only_fastq_file, bcool_inter_only_fastq_file, frequency_file)
                os.system("rm %s" % original_bcool_inter_only_fastq_file)
                os.system("rm %s" % bcool_inter_only_fastq_file)
        ######################################################################
        if new_sequences_set:
            keep_new_seq_bcool_name_lst = []
            for new_seq in new_sequences_set:
                bcool_names = bcool_seq2name_dict[seq]
                temp_bcool_name_lst = []
                for name in bcool_names:
                    ori_seq = str(original_records[name].seq)
                    if editdistance.eval(new_seq, ori_seq) <= 3:
                        temp_bcool_name_lst.append(name)
                if len(temp_bcool_name_lst) >= 2:
                    keep_new_seq_bcool_name_lst.extend(temp_bcool_name_lst)
                del bcool_names, temp_bcool_name_lst

            len2 = len(keep_new_seq_bcool_name_lst)
            if len2 and self.iso_change_detail:
                # save these frquency changing reads before and after correction
                original_bcool_new_seq_only_fastq_file = prefix + '_original_bcool_new_seq_only.' + ori_file_type
                extract_records(self.output_dir, keep_new_seq_bcool_name_lst, f_original, original_bcool_new_seq_only_fastq_file)

                bcool_new_seq_only_fastq_file = prefix + '_bcool_new_seq_only.' + bcool_file_type
                extract_records(self.output_dir, keep_new_seq_bcool_name_lst, f_bcool, bcool_new_seq_only_fastq_file)
                
                with open(frequency_file, 'a') as f:
                    f.write('===========================================')
                    f.write('\n')
                    f.write('Newly generated sequences: ')
                    f.write('\n')
                    f.write('===========================================')
                self.freqency_seq_extraction(original_bcool_new_seq_only_fastq_file, bcool_new_seq_only_fastq_file, frequency_file)
                os.system("rm %s" % original_bcool_new_seq_only_fastq_file)
                os.system("rm %s" % bcool_new_seq_only_fastq_file)

        ##################################################################################################################
        corrected_isolates = self.output_dir + 'corrected_isolates.' + ori_file_type
        keep_correct_f = prefix + '_keep_corrected_records.' + ori_file_type 
        no_change_fastq_file = prefix + 'no_change_isolates.' + ori_file_type
        
        bcool_records, _ = parse_data_index(f_bcool)
        total_name_lst = list(bcool_records)

        if new_sequences_set and intersect_sequences_set:
            keep_bcool_name_lst = list(set(keep_new_seq_bcool_name_lst).union(set(keep_inter_bcool_name_lst)))
        elif new_sequences_set:
            keep_bcool_name_lst = keep_new_seq_bcool_name_lst
        elif intersect_sequences_set:
            keep_bcool_name_lst = keep_inter_bcool_name_lst
        else:
            keep_bcool_name_lst = []
        
        del intersect_sequences_set, new_sequences_set, bcool_records

        if len(keep_bcool_name_lst):
            if bcool_file_type != ori_file_type:
                # self.bcool_seqs(self.num_workers, keep_bcool_name_lst, f_bcool, f_original, keep_correct_f) 
                self.bcool_seqs(keep_bcool_name_lst, f_bcool, f_original, keep_correct_f)  
            else:    
                extract_records(self.output_dir, keep_bcool_name_lst, f_bcool, keep_correct_f)

            rest_name_lst = list(set(total_name_lst) - set(keep_bcool_name_lst))
            if len(rest_name_lst):
                extract_records(self.output_dir, rest_name_lst, f_original, no_change_fastq_file)

                self.logger.debug("rest_name_lst:{}".format(len(rest_name_lst)))  
        corrected_isolates = self.output_dir + 'corrected_isolates.' + ori_file_type
        if os.path.exists(keep_correct_f):
            os.system("cat %s %s > %s" % (keep_correct_f, no_change_fastq_file, corrected_isolates))
            os.system("rm %s" % keep_correct_f)
            os.system("rm %s" % no_change_fastq_file)          
        else:   
            return self.isolates
        return corrected_isolates
'''
    def covert_fq2fa(self, fin, fout):
        # records_dict, _ = parse_data_dict(fin)
        id2seq_desc = {}
        records, _ = parse_data_index(fin)
        name_lst = list(records)
        for name in name_lst:
            seq = str(records[name].seq)
            desc = str(records[name].description)
            id2seq_desc[name] = [seq, desc]
        del records
        fa_records = []
        name_len = len(name_lst)
        with WorkerPool(self.num_workers, shared_objects=id2seq_desc, start_method='fork') as pool:
            with tqdm(total=name_len, desc=self.logger.info("Convert fq to fa"), miniters=int(name_len/self.min_iters)) as pbar:   
                for tmp_rec in pool.imap(self.fq2fa, name_lst):
                    fa_records.append(tmp_rec)
                    pbar.update() 
        with open(fout, "w") as handle:
            SeqIO.write(fa_records, handle, 'fasta')
        del id2seq_desc, name_lst, fa_records, name_len
        return

    def fq2fa(self, shared_objects, name):
        return SeqRecord(shared_objects[name][0], id=name, description=shared_objects[name][1])
    def bcool_seq(self, shared_objects, name):
        bcool_records_dict, ori_records_dict = shared_objects
        return SeqRecord(bcool_records_dict[name].seq, id=ori_records_dict[name].id, description=ori_records_dict[name].description, letter_annotations=ori_records_dict[name].letter_annotations)
'''



# -*- coding: utf-8 -*-
# @Author: Pengyao Ping
# @Date:   2023-01-19 10:56:38
# @Last Modified by:   Pengyao Ping
# @Last Modified time: 2023-05-19 10:55:32

import configparser
import os

class Config(object):
    def __init__(self, config_file, logger):
        conf = configparser.ConfigParser()
        base_path = os.getcwd()
        if config_file:
            conf.read(config_file)
            # path
            if conf.has_option("Paths", "result_dir"):
                self.result_dir = conf.get("Paths", "result_dir")
            else:
                self.result_dir = os.path.join(base_path, 'result/')
                
            # input
            if conf.has_option("SourceInputData", "input_file"):
                self.input_file = conf.get("SourceInputData", "input_file")      

            if conf.has_option("SourceInputData", "ground_truth_data"):
                self.ground_truth_data = conf.get("SourceInputData", "ground_truth_data")
            else:
                self.ground_truth_data = None   
                
            if conf.has_option("SourceInputData", "correct_data"):
                self.correct_data = conf.get("SourceInputData", "correct_data")

            # setting model parameters
            if conf.has_option("General", "num_workers"):
                self.num_workers = conf.getint("General", "num_workers")
            else:
                self.num_workers = -1
            if conf.has_option("General", "chunks_num"):
                self.chunks_num = conf.getint("General", "chunks_num")
            else:   
                self.chunks_num = 100
            if conf.has_option("General", "verbose"):
                self.verbose = conf.getboolean("General", "verbose")
            else:
                self.verbose = False
            if conf.has_option("General", "min_iters"):
                self.min_iters = conf.getint("General", "min_iters")
            else:
                self.min_iters = 1000
            if conf.has_option("General", "iso_change_detail"):
                self.iso_change_detail = conf.getboolean("General", "iso_change_detail")
            else:
                self.iso_change_detail = False
            if conf.has_option("General", "top_n"):
                self.top_n = conf.getint("General", "top_n")
            else:
                self.top_n = 100 
            if conf.has_option("General", "min_read_len"):
                self.min_read_len = conf.getint("General", "min_read_len")
            else:
                self.min_read_len = 30 
            # when the number of negative samples larger than preseting threshold, noise2read will downsample negative samples for training
            if conf.has_option("General", "negative_sample_num"):
                self.negative_sample_num = conf.getint("General", "negative_sample_num")
            else:
                self.negative_sample_num = 300000 # when 
            # if conf.has_option("General", "over_sampling"):
            #     self.over_sampling = conf.getboolean("General", "over_sampling")
            # else:
            #     self.over_sampling = True
            # GraphSetup
            if conf.has_option("GraphSetup", "high_freq_thre"):
                self.high_freq_thre = conf.getint("GraphSetup", "high_freq_thre")
            else:
                self.high_freq_thre = 4

            if conf.has_option("GraphSetup", "max_error_freq"):
                self.max_error_freq = conf.getint("GraphSetup", "max_error_freq")
            else:
                self.max_error_freq = 4

            if conf.has_option("GraphSetup", "save_graph"):
                self.save_graph = conf.getboolean("GraphSetup", "save_graph")
            else:
                self.save_graph = False

            if conf.has_option("GraphSetup", "graph_visualization"):
                self.graph_visualization = conf.getboolean("GraphSetup", "graph_visualization")
            else:
                self.graph_visualization = False

            if conf.has_option("GraphSetup", "drawing_graph_num"):
                self.drawing_graph_num = conf.getint("GraphSetup", "drawing_graph_num")
            else:
                self.drawing_graph_num = 50   
            # EmbeddingSetup
            if conf.has_option("EmbeddingSetup", "entropy_kmer"):
                self.entropy_kmer = conf.getint("EmbeddingSetup", "entropy_kmer")
            else:
                self.entropy_kmer = 3  
            if conf.has_option("EmbeddingSetup", "entropy_q"):
                self.entropy_q = conf.getint("EmbeddingSetup", "entropy_q")
            else:
                self.entropy_q = 2  
            if conf.has_option("EmbeddingSetup", "kmer_freq"):
                self.kmer_freq = conf.getint("EmbeddingSetup", "kmer_freq")
            else:
                self.kmer_freq = 3  
            if conf.has_option("EmbeddingSetup", "read_type"):
                self.read_type = conf.get("EmbeddingSetup", "read_type")
            else:
                self.read_type = 'DNA'

            # AmbiguousSetup
            # if conf.has_option("AmbiguousSetup", "ambiguous_error_node_degree"):
            #     self.ambiguous_error_node_degree = conf.getint("AmbiguousSetup", "ambiguous_error_node_degree")
            # else:
            #     self.ambiguous_error_node_degree = 4 # default 
            if conf.has_option("AmbiguousSetup", "high_ambiguous"):
                self.high_ambiguous = conf.getboolean("AmbiguousSetup", "high_ambiguous")
            else:
                self.high_ambiguous = True        
            if conf.has_option("AmbiguousSetup", "proba_deviation"):
                self.proba_deviation = conf.getfloat("AmbiguousSetup", "proba_deviation")
            else:
                self.proba_deviation = 0.95 # default 

            # ModelTuningSetup
            if conf.has_option("ModelTuningSetup", "n_trials"):
                self.n_trials = conf.getint("ModelTuningSetup", "n_trials")
            else:
                self.n_trials = 30
                
            if conf.has_option("ModelTuningSetup", "n_estimators"):
                self.n_estimators = conf.getint("ModelTuningSetup", "n_estimators")
            else:
                self.n_estimators = 400

            if conf.has_option("ModelTuningSetup", "test_size"):
                self.test_size = conf.getfloat("ModelTuningSetup", "test_size")
            else:
                self.test_size = 0.1 # default        
            if conf.has_option("ModelTuningSetup", "random_state"):
                self.random_state = conf.getint("ModelTuningSetup", "random_state")
            else:
                self.random_state = 42 # default  

            if conf.has_option("ModelTuningSetup", "tree_method"):
                self.tree_method = conf.get("ModelTuningSetup", "tree_method")
            else:
                self.tree_method = 'auto'            

            if conf.has_option("ModelTuningSetup", "learning_rate_min"):
                self.learning_rate_min = conf.getfloat("ModelTuningSetup", "learning_rate_min")
            else:
                self.learning_rate_min = 1e-3 # default     
            if conf.has_option("ModelTuningSetup", "learning_rate_max"):
                self.learning_rate_max = conf.getfloat("ModelTuningSetup", "learning_rate_max")
            else:
                self.learning_rate_max = 1e-1 # default 

            if conf.has_option("ModelTuningSetup", "max_depth_min"):
                self.max_depth_min = conf.getint("ModelTuningSetup", "max_depth_min")
            else:
                self.max_depth_min = 3 # default     
            if conf.has_option("ModelTuningSetup", "max_depth_max"):
                self.max_depth_max = conf.getint("ModelTuningSetup", "max_depth_max")
            else:
                self.max_depth_max = 15 # default     
            if conf.has_option("ModelTuningSetup", "max_depth_step"):
                self.max_depth_step = conf.getint("ModelTuningSetup", "max_depth_step")
            else:
                self.max_depth_step = 1 # default 

            if conf.has_option("ModelTuningSetup", "num_boost_round_min"):
                self.num_boost_round_min = conf.getint("ModelTuningSetup", "num_boost_round_min")
            else:
                self.num_boost_round_min = 200 # default     
            if conf.has_option("ModelTuningSetup", "num_boost_round_max"):
                self.num_boost_round_max = conf.getint("ModelTuningSetup", "num_boost_round_max")
            else:
                self.num_boost_round_max = 300 # default     
            if conf.has_option("ModelTuningSetup", "num_boost_round_step"):
                self.num_boost_round_step = conf.getint("ModelTuningSetup", "num_boost_round_step")
            else:
                self.num_boost_round_step = 10 # default 

            if conf.has_option("ModelTuningSetup", "subsample_min"):
                self.subsample_min = conf.getfloat("ModelTuningSetup", "subsample_min")
            else:
                self.subsample_min = 0.8 # default     
            if conf.has_option("ModelTuningSetup", "subsample_max"):
                self.subsample_max = conf.getfloat("ModelTuningSetup", "subsample_max")
            else:
                self.subsample_max = 1.0 # default     

            if conf.has_option("ModelTuningSetup", "colsample_bytree_min"):
                self.colsample_bytree_min = conf.getfloat("ModelTuningSetup", "colsample_bytree_min")
            else:
                self.colsample_bytree_min = 0.8 # default     
            if conf.has_option("ModelTuningSetup", "colsample_bytree_max"):
                self.colsample_bytree_max = conf.getfloat("ModelTuningSetup", "colsample_bytree_max")
            else:
                self.colsample_bytree_max = 1.0 # default     
            
            if conf.has_option("ModelTuningSetup", "verbose_eval"):
                self.verbose_eval = conf.getboolean("ModelTuningSetup", "verbose_eval")
            else:
                self.verbose_eval = False
                
            # xgboostclassifier seed
            if conf.has_option("ModelTuningSetup", "xgboost_seed"):
                self.xgboost_seed = conf.getint("ModelTuningSetup", "xgboost_seed")
            else:
                self.xgboost_seed = 42 # default 

            if conf.has_option("ModelTuningSetup", "optuna_seed"):
                self.optuna_seed = conf.getint("ModelTuningSetup", "optuna_seed")
            else:
                self.optuna_seed = 42 # default 

            if conf.has_option("ModelTuningSetup", "best_accuracy"):
                self.best_accuracy = conf.getfloat("ModelTuningSetup", "best_accuracy")
            else:
                self.best_accuracy = 0.85 # default     

            # real umi
            if conf.has_option("RealUMI", "umi_start"):
                self.umi_start = conf.getint("RealUMI", "umi_start")
            else:
                self.umi_start = 0 # default 
            if conf.has_option("RealUMI", "umi_end"):
                self.umi_end = conf.getint("RealUMI", "umi_end")
            else:
                self.umi_end = 12 # default 
            if conf.has_option("RealUMI", "non_umi_start"):
                self.non_umi_start = conf.getint("RealUMI", "non_umi_start")
            else:
                self.non_umi_start = 24 # default
            # amplicon
            if conf.has_option("Amplicon", "amplicon_low_freq"):
                self.amplicon_low_freq = conf.getint("Amplicon", "amplicon_low_freq")
            else:
                self.amplicon_low_freq = 50 # default             
            if conf.has_option("Amplicon", "amplicon_high_freq"):
                self.amplicon_high_freq = conf.getint("Amplicon", "amplicon_high_freq")
            else:
                self.amplicon_high_freq = 5000 # default   
            if conf.has_option("Amplicon", "amplicon_threshold_proba"):
                self.amplicon_threshold_proba = conf.getfloat("Amplicon", "amplicon_threshold_proba")
            else:
                self.amplicon_threshold_proba = 0.85 # default   
            # if conf.has_option("Amplicon", "amplicon_error_node_degree"):
            #     self.amplicon_error_node_degree = conf.getint("Amplicon", "amplicon_error_node_degree")
            # else:
            #     self.amplicon_error_node_degree = 4 # default   

            # Simulation
            if conf.has_option("Simulation", "min_freq"):
                self.min_freq = conf.getint("Simulation", "min_freq")
            else:
                self.min_freq = 4 # default
            if conf.has_option("Simulation", "min_read_count"):
                self.min_read_count = conf.getint("Simulation", "min_read_count")
            else:
                self.min_read_count = 30 # default
                
            if conf.has_option("Simulation", "substations"):
                self.substations = conf.getboolean("Simulation", "substations")
            else:
                self.substations = True
            if conf.has_option("Simulation", "indels"):
                self.indels = conf.getboolean("Simulation", "indels")
            else:
                self.indels = True

            if conf.has_option("Simulation", "error_rate1"):
                self.error_rate1 = conf.getfloat("Simulation", "error_rate1")
            else:
                self.error_rate1 = 0.03 # default

            if conf.has_option("Simulation", "error_rate2"):
                self.error_rate2 = conf.getfloat("Simulation", "error_rate2")
            else:
                self.error_rate2 = 0.005 # default

            if conf.has_option("Simulation", "sim_random_state"):
                self.sim_random_state = conf.getint("Simulation", "sim_random_state")
            else:
                self.sim_random_state = 42 # default
                
            # # Evaluation
            # if conf.has_option("Evaluation", "delta"):
            #     self.delta = conf.getint("Evaluation", "delta")
            # else:
            #     self.delta = 10 # default

        if config_file == None:
            # path
            self.result_dir = os.path.join(base_path, 'result/')
            # input
            self.ground_truth_data = None  
            # general
            self.num_workers = -1 
            self.chunks_num = 100
            self.min_iters = 1000
            self.verbose = True 
            self.iso_change_detail = False     
            self.top_n = 100      
            # self.over_sampling = True 
            self.negative_sample_num = 300000
            self.min_read_len = 30

            # GraphSetup
            self.high_freq_thre = 4
            self.max_error_freq = 4
            self.save_graph = False
            self.graph_visualization = False
            self.drawing_graph_num = 50   

            # EmbeddingSetup
            self.entropy_kmer = 3
            self.entropy_q = 2
            self.kmer_freq = 3
            self.read_type = 'DNA'

            # AmbiguousSetup
            self.high_ambiguous = True 
            # high ambiguous predict probability difference
            self.proba_deviation = 0.95      # for base editing and lncRNA quant datasets 0.75 others 0.6 
            # self.ambiguous_error_node_degree = 4   # for base editing and lncRNA quant datasets 3 others 4 

            # ModelTuningSetup
            self.n_trials = 30
            self.n_estimators = 400
            self.test_size = 0.1 # default 
            # random state for SMOTE and train_test_split       
            self.random_state = 42 # default  
            self.tree_method = 'auto'
            self.learning_rate_min = 1e-3 # default     
            self.learning_rate_max = 1e-1 # default 
            self.max_depth_min = 3 # default     
            self.max_depth_max = 15 # default     
            self.max_depth_step = 1 # default 
            self.num_boost_round_min = 200 # default     
            self.num_boost_round_max = 300 # default     
            self.num_boost_round_step = 10 # default 
            self.subsample_min = 0.8 # default     
            self.subsample_max = 1 # default     
            self.colsample_bytree_min = 0.8 # default     
            self.colsample_bytree_max = 1 # default     
            self.verbose_eval = False
            # xgboostclassifier seed
            self.xgboost_seed = 42 # default 
            self.optuna_seed = 42
            # optuna best trial accuracy
            self.best_accuracy = 0.85

            # real umi
            self.umi_start = 0
            self.umi_end = 12
            self.non_umi_start = 24

            #amplicon
            self.amplicon_low_freq = 50
            self.amplicon_high_freq = 5000
            self.amplicon_threshold_proba = 0.85
            # self.amplicon_error_node_degree = 4

            # simulation
            self.min_freq = 4
            self.min_read_count = 30
            self.substations = True
            self.indels = True
            self.error_rate1 = 0.03
            self.error_rate2 = 0.005
            self.sim_random_state = 42

            # # Evaluation
            # self.delta = 1

            # # coverage
            # self.library_layout = 'PE'
            # self.Alignment = '--local' # '--end-to-end'
            
            



# -*- coding: utf-8 -*-
# @Author: Pengyao Ping
# @Date:   2023-02-16 11:01:06
# @Last Modified by:   Pengyao Ping
# @Last Modified time: 2023-05-17 10:24:07

import collections
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
import os
from noise2read.utils import *
from noise2read.data_generation import DataGneration
from noise2read.error_orrection import ErrorCorrection
# from noise2read.data_analysis import DataAnalysis
import editdistance
from tqdm import tqdm
from Bio.Seq import Seq

class DataProcessing():
    # def __init__(self, logger, num_workers, output_dir, umi_start=None, umi_end=None, non_umi_start=None):
    def __init__(self, logger, config):
        self.logger = logger
        self.config = config
        
        if os.path.exists(self.config.result_dir):
            self.logger.info("Directory '% s' already exists" % self.config.result_dir)
        else:
            os.makedirs(self.config.result_dir)
        if os.path.exists(self.config.result_dir + 'raw/'):
            self.logger.info("Directory '% s' already exists" % self.config.result_dir + 'raw/')
        else:
            os.makedirs(self.config.result_dir + 'raw/')       
        if os.path.exists(self.config.result_dir + 'true/'):
            self.logger.info("Directory '% s' already exists" % self.config.result_dir + 'true/')
        else:
            os.makedirs(self.config.result_dir + 'true/') 
               
    def extract_umis(self, original_data, umi_start, umi_end, non_umi_start):
        record_iterator, ff_type = parse_data(original_data) 
        self.logger.info("Read Data")
        umi_records = []
        non_umi_records = []

        for item in tqdm(record_iterator):
            seq = str(item.seq)
            if 'N' not in seq:
                umi_qual = {}
                non_umi_qual = {}
                if ff_type == 'fastq' or ff_type == 'fq' or ff_type == 'fastq.gz' or ff_type == 'fq.gz':
                    val = item.letter_annotations['phred_quality']            
                umi_desc = "length=" + str(umi_end-umi_start)
                non_umi_desc = "length=" + str(len(seq) - non_umi_start)
                if ff_type == "fastq": 
                    umi_qual['phred_quality'] = val[umi_start:umi_end]
                    non_umi_qual['phred_quality'] = val[non_umi_start:]
                    umi_rec = SeqRecord(Seq(seq[umi_start:umi_end]), id=item.id, description=umi_desc, letter_annotations=umi_qual)   
                    non_umi_rec = SeqRecord(Seq(seq[non_umi_start:]), id=item.id, description=non_umi_desc, letter_annotations=non_umi_qual)   
                elif ff_type == "fasta":
                    umi_rec = SeqRecord(Seq(seq[umi_start:umi_end]), id=item.id, description=umi_desc)        
                    non_umi_rec = SeqRecord(Seq(seq[non_umi_start:]), id=item.id, description=non_umi_desc)  
                umi_records.append(umi_rec)
                non_umi_records.append(non_umi_rec)        
        self.logger.info("Rewrite umi and non-umi Data")
        raw_base_name = original_data.split('/')[-1]
        umi_raw_dataset = self.output_dir + "umi_" + raw_base_name
        non_umi_raw_dataset = self.output_dir + "non_umi_" + raw_base_name
        with open(umi_raw_dataset, "w") as handle:
            SeqIO.write(umi_records, handle, ff_type)        
        with open(non_umi_raw_dataset, "w") as handle:
            SeqIO.write(non_umi_records, handle, ff_type) 
        self.logger.info("Split umi and non-umis completed.") 
        return umi_raw_dataset, non_umi_raw_dataset

    def real_umi_data(self, original_data):
        # step1: extract umis to a sperate fastq file
        umi_raw_dataset, non_umi_raw_dataset = self.extract_umis(original_data, self.config.umi_start, self.config.umi_end, self.config.non_umi_start)
        # step2: correct umi errors
        self.config.input_file = umi_raw_dataset
        DG = DataGneration(self.logger, self.config)
        # DG = DataGneration(
        #     self.logger,
        #     self.config.num_workers,
        #     umi_raw_dataset, 
        #     self.config.output_dir,
        #     high_freq_thre = self.config.high_freq_thre,
        #     max_error_freq = self.config.max_error_freq,
        #     save_graph = self.config.save_graph,
        #     graph_visualization = self.config.graph_visualization,
        #     drawing_graph_num = self.config.drawing_graph_num,
        #     high_ambiguous=self.config.high_ambiguous, 
        #     verbose=self.config.verbose
        #     )
        genuine_df = DG.extract_umi_genuine_errs()
        # ##############################################################
        EC = ErrorCorrection(self.logger, self.config)
        # EC = ErrorCorrection(
        #     self.logger,
        #     self.num_workers,
        #     umi_raw_dataset,
        #     self.output_dir,
        #     read_max_len = self.umi_end - self.umi_start,
        #     entropy_kmer=3, 
        #     entropy_q=2, 
        #     kmer_freq=3,
        #     read_type="DNA",
        #     iso_change_detail=False,
        #     min_iters=100)
        corrected_file = EC.umi_correction(umi_raw_dataset, genuine_df)

        # base_name = umi_raw_dataset.split("/")[-1].split(".fastq")[0]
        # step3: use umi and real dataset to generate raw and true dataset
        raw_file, true_file = self.raw_true_umi(corrected_file, non_umi_raw_dataset)
        return raw_file, true_file

    def raw_true_umi(self, correct_umi_f, non_umi_f):
        umi2id = {}
        umi_record_iterator, umi_f_type = parse_data(correct_umi_f)
        for item in umi_record_iterator:
            umi = str(item.seq)
            umi2id.setdefault(umi, []).append(str(item.id))

        id2read, ori_file_type = parse_data_dict(non_umi_f)
        raw_records = []
        true_records = []
        for umi in umi2id:
            ids = umi2id[umi]
            # n = 0
            tmp_raw_rec = []
            tmp_true_rec = []
            if len(ids) >= 10:
                cur_seqs = []
                for id in ids:
                    cur_seqs.append(str(id2read[id].seq))
                seq2count = collections.Counter(cur_seqs)
                max_key = max(seq2count, key=seq2count.get)    
                total_count = sum(list(seq2count.values()))
                max_key_count = seq2count[max_key]
                if max_key_count >= (total_count-max_key_count):     
                    for i in ids:
                        cur_seq = str(id2read[i].seq)
                        dis = editdistance.eval(max_key, cur_seq)
                        desc = "umi:" + str(umi) + "//" + str(id2read[i].description)
                        if dis == 0:
                            # n += 1
                            if ori_file_type == "fastq":
                                rec = SeqRecord(Seq(cur_seq), id=i, description=desc, letter_annotations=id2read[i].letter_annotations)
                            else:
                                rec = SeqRecord(Seq(cur_seq), id=i, description=desc)
                            # raw_records.append(rec)
                            # true_records.append(rec)
                            tmp_raw_rec.append(rec)
                            tmp_true_rec.append(rec)
                        elif dis == 1 or dis == 2:
                            # n += 1
                            if ori_file_type == "fastq":
                                raw_rec = SeqRecord(Seq(cur_seq), id=i, description=desc, letter_annotations=id2read[i].letter_annotations)
                                true_rec = SeqRecord(Seq(max_key), id=i, description=desc, letter_annotations=id2read[i].letter_annotations)
                            else:
                                raw_rec = SeqRecord(Seq(cur_seq), id=i, description=desc)
                                true_rec = SeqRecord(Seq(max_key), id=i, description=desc)
                            # raw_records.append(raw_rec)
                            # true_records.append(true_rec)
                            tmp_raw_rec.append(raw_rec)
                            tmp_true_rec.append(true_rec)
                    # if n >= 10:
                    raw_records.extend(tmp_raw_rec)
                    true_records.extend(tmp_true_rec)                    
        self.logger.info("Write raw and true data to file")
        raw_file = self.output_dir + "raw/" + non_umi_f.split('/')[-1]
        true_file = self.output_dir + "true/" + non_umi_f.split('/')[-1]
        with open(raw_file, "w") as handle:
            SeqIO.write(raw_records, handle, ori_file_type)        
        with open(true_file, "w") as handle:
            SeqIO.write(true_records, handle, ori_file_type) 
        self.logger.info("Real umi data processing completed.")                 

        return raw_file, true_file

    def write_mimic_umis(self, raw_dataset, true_dataset):
        true_records_dict, true_file_type = parse_data_dict(true_dataset)

        true_seqs_lst = []
        # raw_seqs_lst = []
        for name in true_records_dict:
            true_seq = str(true_records_dict[name].seq)
            # raw_seq = str(error_records[name].seq)
            true_seqs_lst.append(true_seq)     
            # raw_seqs_lst.append(raw_seq)
        # raw_read2count = collections.Counter(raw_seqs_lst)
        true_read2count = collections.Counter(true_seqs_lst)
        # write the same read the same id in true dataset, these ids also write to raw dataset
        true_read2id = {}
        i = 1
        for read in true_read2count:
            true_read2id[read] = i
            i += 1
        true_records = []
        raw_records = []
        raw_records_dict, raw_file_type = parse_data_dict(raw_dataset)
        self.logger.info("Read Data")
        for name in tqdm(true_records_dict):
            true_seq = str(true_records_dict[name].seq)
            true_des = "umi:" + str(true_read2id[true_seq]) + "//" + str(true_records_dict[name].description)
            raw_des = "umi:" + str(true_read2id[true_seq]) + "//" + str(raw_records_dict[name].description)
            if true_file_type == "fastq": 
                true_rec = SeqRecord(true_records_dict[name].seq, id=true_records_dict[name].id, description=true_des, letter_annotations=true_records_dict[name].letter_annotations)   
                raw_rec = SeqRecord(raw_records_dict[name].seq, id=raw_records_dict[name].id, description=raw_des, letter_annotations=raw_records_dict[name].letter_annotations)   
            elif true_file_type == "fasta":
                true_rec = SeqRecord(true_records_dict[name].seq, id=true_records_dict[name].id, description=true_des)        
                raw_rec = SeqRecord(raw_records_dict[name].seq, id=raw_records_dict[name].id, description=raw_des)  
            raw_records.append(raw_rec)
            true_records.append(true_rec)
        raw_base_name = raw_dataset.split('/')[-1]
        true_base_name = true_dataset.split('/')[-1]
        umi_raw_dataset = self.output_dir + "raw/" + "umi_" + raw_base_name
        umi_true_dataset = self.output_dir + "true/" + "umi_"  + true_base_name
        self.logger.info("Rewrite Data")
        with open(umi_raw_dataset, "w") as handle:
            SeqIO.write(raw_records, handle, raw_file_type)        
        with open(umi_true_dataset, "w") as handle:
            SeqIO.write(true_records, handle, true_file_type) 
        self.logger.info("Mimic_umi preprocessing completed.") 
        return umi_raw_dataset, umi_true_dataset


# -*- coding: utf-8 -*-
# @Author: Pengyao Ping
# @Date:   2023-01-30 09:35:18
# @Last Modified by:   Pengyao Ping
# @Last Modified time: 2023-04-11 17:09:30

from collections import Counter
import collections
import math
from Bio import SeqIO
import os
import editdistance
import xlsxwriter
from noise2read.utils import *
from numpy.linalg import norm
from numpy import array
from numpy.fft import fft,rfft
# import numpy as np
# import antropy as ant
import cmath
import numpy as np
import math
import matplotlib.pyplot as plt
import seaborn as sns
import matplotlib as mpl
from matplotlib.colors import ListedColormap, LinearSegmentedColormap
from random import shuffle
from tqdm import tqdm
from mpire import WorkerPool

class DataAnalysis():
    """
    A class to evaluate an error correction method.
    """
    def __init__(self, logger, config):
        """
        initialize the DataAnalysis class

        Args:
            logger (class): customized logging
            config (class): parameters setting using configparser
        """
        self.logger = logger
        self.config = config
        self.logger.info("-------------------------------------------------------------")
        self.logger.info(f'Raw dataset: {config.input_file}')
        self.logger.info(f'Correct dataset: {config.correct_data}')
        self.logger.info(f'Ground truth dataset: {config.ground_truth_data}')
        if os.path.exists(self.config.result_dir):
            self.logger.info("Directory '% s' already exists" % self.config.result_dir)
            # for f in os.listdir(self.config.result_dir):
            #     os.remove(os.path.join(self.config.result_dir, f))
        else:
            os.makedirs(self.config.result_dir)
            self.logger.info("Directory '% s' created" % self.config.result_dir)
        bases = config.correct_data.split('/')[-1]
        self.prefix = bases.split('.' + parse_file_type(config.correct_data))[0]
        self.result_file = os.path.join(config.result_dir, self.prefix + '.xlsx')
        self.workbook_flie = xlsxwriter.Workbook(self.result_file)
        
    def percentage(self, part, whole):
        """
        percentage calculation

        Args:
            part (int): _description_
            whole (int): _description_

        Returns:
            str: percentage, if whole equals 0, return 'Not exist in raw'
        """
        if whole != 0:
            Percentage = 100 * float(part)/float(whole)
            return str(Percentage) + "%"
        else:
            return str('Not exist in raw')
        
    def ranking(self, num, num_lst):
        """
        get read count's ranking

        Args:
            num (int): read count
            num_lst (list): a list of read counts

        Returns:
            int: ranking
        """
        for n in num_lst:
            if num == n:
                return num_lst.index(n) + 1

    def evaluation(self):
        """
        Invoke the function to run the whole evaluation process, and write the result to files. 
        """
        # read the input using SeqIO
        raw_record_iterator, raw_file_tye = parse_data(self.config.input_file)
        correct_record_iterator, correct_file_tye = parse_data(self.config.correct_data)
        raw_seqs = []
        correct_seqs = []

        self.raw_len2seqs_dict = {}
        self.raw_len_lst = []
        total_reads_num = 0
        
        raw_record_dict = {}
        correct_record_dict = {}
        id_lst = []
        for raw_item, correct_item in zip(raw_record_iterator, correct_record_iterator):
            raw_id = str(raw_item.id)
            raw_seq = str(raw_item.seq)

            cor_id = str(correct_item.id)
            cor_seq = str(correct_item.seq)
            
            raw_seqs.append(raw_seq)
            correct_seqs.append(cor_seq)
            
            ll = len(raw_seq)
            self.raw_len_lst.append(ll)
            self.raw_len2seqs_dict.setdefault(ll, []).append(raw_seq)
            total_reads_num += 1
            
            raw_record_dict[raw_id] = raw_seq
            correct_record_dict[cor_id] = cor_seq
            
            id_lst.append(raw_id)

        corrected_reads_num = 0

        for item in tqdm(id_lst):
            ori_read = raw_record_dict[item]
            cor_read = correct_record_dict[item]
            if str(cor_read) != str(ori_read):
                corrected_reads_num += 1
                
        raw_read2count = Counter(raw_seqs)
        correct_read2count = Counter(correct_seqs)

        raw_read2count_val_lst = list(raw_read2count.values())
        raw_read2count_val_lst.sort(reverse=True)

        correct_read2count_val_lst = list(correct_read2count.values())
        correct_read2count_val_lst.sort(reverse=True)

        raw_counts_lst = raw_read2count_val_lst[ : self.config.top_n * 10]
        correct_counts_lst = correct_read2count_val_lst[ : self.config.top_n * 10]
        
        # self.workbook_flie = xlsxwriter.Workbook(self.result_file)
        worksheet = self.workbook_flie.add_worksheet('Read Count')
        # Use the worksheet object to write
        # data via the write() method.
        worksheet.write('A1', 'Read')
        worksheet.write('B1', 'Count after Correction')
        worksheet.write('C1', 'Count before Correction')
        worksheet.write('D1', 'Change in percentage')
        worksheet.write('E1', 'Rank after Correction')
        worksheet.write('F1', 'Rank before Correction')

        row = 1
        col = 0
        for read, count in correct_read2count.most_common(self.config.top_n):
            worksheet.write(row, col, read)
            worksheet.write(row, col + 1, count)
            
            raw_read_count = raw_read2count[read]
            worksheet.write(row, col + 2, raw_read_count)
            abundance_percentage = self.percentage(count - raw_read_count, raw_read_count)
            worksheet.write(row, col + 3, abundance_percentage)

            correct_rank = self.ranking(count, correct_counts_lst)
            worksheet.write(row, col + 4, correct_rank)
            raw_rank = self.ranking(raw_read_count, raw_read2count_val_lst)
            worksheet.write(row, col + 5, raw_rank)
            row += 1

        self.logger.debug(f'Ground truth: {self.config.ground_truth_data}')
        if self.config.ground_truth_data:
            self.evaluation_with_groundtruth()
        self.logger.info("corrected {} reads out of {} ({:.6f}) reads".format(corrected_reads_num, total_reads_num, corrected_reads_num/total_reads_num))
        raw_unique_num = len(raw_read2count)
        correct_unique_num = len(correct_read2count)
        self.logger.info("Unique reads decreased by {} from {} to {} ({:.6f}) reads".format(raw_unique_num - correct_unique_num, raw_unique_num, correct_unique_num, (raw_unique_num - correct_unique_num)/raw_unique_num))
        # no ground truth entropy
        self.evaluation_without_groundtruth(raw_read2count, correct_read2count, total_reads_num)
        return

    # rely on read frequency instead of sequecing id
    def evaluation_with_groundtruth(self):
        """
        Evaluating an error correction method using raw, true and corrected data based on read frequency instead of sequecing id
        """
        read_level = {'tp':0, 'fp':0, 'tn':0, 'fn':0}
        base_level = {'tp':0, 'fp':0, 'tn':0, 'fn':0}
        
        true_records, true_f_tye = parse_data_index(self.config.ground_truth_data)
        error_records, err_f_tye = parse_data_index(self.config.input_file)
        correct_records, cor_f_tye = parse_data_index(self.config.correct_data)

        total_reads_num = 0
        # for calculating entropy
        correct_errfree_seqs = []
        correct_err_seqs = []
        raw_errfreee_seqs = []
        raw_err_seqs = []

        true_seqs_lst = []
        raw_seqs_lst = []
        correct_seqs_lst = []
        true_umi2seqs = {}
        raw_umi2seqs = {}
        correct_umi2seqs = {}

        true_seq2umi = {}
        for name in error_records:
            true_seq = str(true_records[name].seq)
            raw_seq = str(error_records[name].seq)
            correct_seq = str(correct_records[name].seq) 

            umi_base = str(true_records[name].description).split('//')[0]
            umi = umi_base.split(':')[1]
            true_umi2seqs.setdefault(umi, []).append(true_seq)
            raw_umi2seqs.setdefault(umi, []).append(raw_seq)
            correct_umi2seqs.setdefault(umi, []).append(correct_seq)

            true_seq2umi.setdefault(true_seq, []).append(umi)

            true_seqs_lst.append(true_seq)     
            raw_seqs_lst.append(raw_seq)
            correct_seqs_lst.append(correct_seq)

        raw_read2count = collections.Counter(raw_seqs_lst)
        true_read2count = collections.Counter(true_seqs_lst)
        correct_read2count = collections.Counter(correct_seqs_lst)

        num = 0
        read_num = 0
        for seq in true_seq2umi:
            unique_umi = set(true_seq2umi[seq])
            if len(unique_umi) > 1:
                num += 1
                read_num += correct_read2count[seq]
                # print(unique_umi, )
        print(num, read_num)
############################################################################################
        # merge same sequences with different umis
        true_umis2seqs = {}
        raw_umis2seqs = {}
        correct_umis2seqs = {}
        for seq in true_seq2umi:
            unique_umi = list(set(true_seq2umi[seq]))
            umi_num = len(unique_umi)
            umi = tuple(unique_umi)
            if umi_num == 1:
                tmp_umi = unique_umi[0]
                true_umis2seqs[umi] = true_umi2seqs[tmp_umi]
                raw_umis2seqs[umi] = raw_umi2seqs[tmp_umi]
                correct_umis2seqs[umi] = correct_umi2seqs[tmp_umi]   
            else:
                for tmp_umi in unique_umi:
                    true_umis2seqs.setdefault(umi, []).extend(true_umi2seqs[tmp_umi])   
                    raw_umis2seqs.setdefault(umi, []).extend(raw_umi2seqs[tmp_umi])
                    correct_umis2seqs.setdefault(umi, []).extend(correct_umi2seqs[tmp_umi])      


############################################################################################
        fn_lst = []
        positive_fp_lst = []
        # ii = 0
        original_high_ambiguous = []
        # for umi in true_umi2seqs:
        #     true_seqs = true_umi2seqs[umi]
        #     raw_seqs = raw_umi2seqs[umi]
        #     correct_seqs = correct_umi2seqs[umi]

        for umi in true_umis2seqs:
            true_seqs = true_umis2seqs[umi]
            raw_seqs = raw_umis2seqs[umi]
            correct_seqs = correct_umis2seqs[umi]

            true_seqs_len = len(true_seqs)
            total_reads_num += true_seqs_len
            # raw base level
            total_positive_bases = 0
            total_negative_bases = 0
            # raw read level
            total_positive_reads = 0
            total_negative_reads = 0

            # correction base level
            cor_positive_bases = 0
            cor_negative_bases = 0
            # correction read level
            cor_positive_reads = 0
            cor_negative_reads = 0
            
            if len(set(true_seqs)) != 1:
                self.logger.exception("UMI(s) contain(s) more than one true sequences.")
            t_seq = true_seqs[0]
            r_seqs = list(set(raw_seqs))
            r_seq2counts = collections.Counter(raw_seqs)
            if len(r_seqs) > 1:
                for r_seq in r_seqs:
                    t_r_dis = editdistance.eval(t_seq, r_seq)
                    # if t_r_dis == 1 or t_r_dis == 2:
                    if t_r_dis == 1:
                        r_seq_count = raw_read2count[r_seq]
                        t_seq_count = raw_read2count[t_seq]
                        if r_seq_count > 5 and t_seq_count > 5:
                            # original_high_ambiguous.append([umi, t_r_dis, t_seq, t_seq_count, r_seq, r_seq_count, r_seq2counts[r_seq]])
                            line = [str(umi), t_r_dis, t_seq, t_seq_count, r_seq, r_seq_count, r_seq2counts[r_seq]]
                            if line not in original_high_ambiguous:
                                original_high_ambiguous.append(line)
                # ii += 1

            for i in range(true_seqs_len):
                raw_seq = raw_seqs[i]
                true_seq = true_seqs[i]
                seq_len = len(true_seq)
                true_raw_dis = editdistance.eval(true_seq, raw_seq)
                # raw base level
                total_positive_bases += true_raw_dis
                total_negative_bases += seq_len - true_raw_dis
                # raw read level
                if true_raw_dis == 0:
                    raw_errfreee_seqs.append(raw_seq)
                    total_negative_reads += 1
                else:
                    raw_err_seqs.append(raw_seq)
                    total_positive_reads += 1
                # after correction
                correct_seq = correct_seqs[i]
                true_cor_dis = editdistance.eval(true_seq, correct_seq)
                # correction base level
                cor_positive_bases += true_cor_dis
                cor_negative_bases += seq_len - true_cor_dis
                # correction read level
                if true_cor_dis == 0:
                    correct_errfree_seqs.append(correct_seq)
                    cor_negative_reads += 1
                else: 
                    correct_err_seqs.append(correct_seq)
                    cor_positive_reads += 1
            # note: for any postive read, if it is modified correctly, tp will increase one. if it is modified incorrectly, it will keep as a fn rather than a fp. Similarly, for any negative read, 
            # base level
            if total_positive_bases >= cor_positive_bases:
                base_level['tp'] += total_positive_bases - cor_positive_bases
                base_level['fn'] += cor_positive_bases  
            if total_negative_bases >= cor_negative_bases:  
                base_level['tn'] += cor_negative_bases
                base_level['fp'] += total_negative_bases - cor_negative_bases

            if total_positive_bases < cor_positive_bases:
                base_level['tp'] += 0
                base_level['fn'] += total_positive_bases
                # self.logger.warning(f'Base-level Evaluation: introduced additional {cor_positive_bases - total_positive_bases} positive bases after correction for UMI {str(umi)}')  
            if total_negative_bases < cor_negative_bases:
                base_level['tn'] += total_negative_bases
                base_level['fp'] += 0    
                # self.logger.warning(f'Base-level Evaluation: introduced additional {cor_negative_bases - total_negative_bases} negative bases after correction for UMI {str(umi)}')   

            # read level
            if total_positive_reads >= cor_positive_reads:
                read_level['tp'] += total_positive_reads - cor_positive_reads
                read_level['fn'] += cor_positive_reads  
            if total_negative_reads >= cor_negative_reads:
                read_level['tn'] += cor_negative_reads
                read_level['fp'] += total_negative_reads - cor_negative_reads  
            if total_positive_reads < cor_positive_reads:
                read_level['tp'] += 0
                read_level['fn'] += total_positive_reads
                # self.logger.warning(f'Read-level Evaluation: introduced additional {cor_positive_reads - total_positive_reads} positive reads after correction for UMI {str(umi)}') 
            if total_negative_reads < cor_negative_reads:
                read_level['tn'] += total_negative_reads
                read_level['fp'] += 0
                # self.logger.warning(f'Read-level Evaluation: introduced additional {cor_negative_reads - total_negative_reads} negative reads after correction for UMI {str(umi)}') 

            ## fn
            raw_positive_seqs = set(raw_seqs) - set(true_seqs)
            untouched_positive_seqs = set(correct_seqs) - set(true_seqs)

            if len(raw_positive_seqs) > 0:
                if len(untouched_positive_seqs) > 0:
                    # true_group = [t_seq, true_read2count[t_seq], raw_read2count[t_seq], correct_read2count[t_seq]]
                    for seq in untouched_positive_seqs:
                        fn_lst.append([str(umi), t_seq, raw_read2count[t_seq], seq, raw_read2count[seq], correct_read2count[t_seq], correct_read2count[seq]])                                        
            # fp
            # elif len(raw_positive_seqs) == 0:
            #     if len(untouched_positive_seqs) > 0:
            #         true_group = [t_seq, true_read2count[t_seq], raw_read2count[t_seq], correct_read2count[t_seq]]
            #         correct_group = []
            #         for item in untouched_positive_seqs:
            #             correct_group.append([item, true_read2count[item], raw_read2count[item], correct_read2count[item]])  
            #         positive_fp_lst.append([umi, str(true_group), str(correct_group)]) 

            # if len(raw_positive_seqs) > 0:
            #     if len(untouched_positive_seqs) > 0:
            #         intersec_seqs = untouched_positive_seqs & raw_positive_seqs
            #         raw_group = []
            #         if len(intersec_seqs) > 0:
            #             for item in intersec_seqs:
            #                 raw_group.append([item, true_read2count[item], raw_read2count[item], correct_read2count[item]])                    
            #         correct_group = []
            #         true_group = [t_seq, true_read2count[t_seq], raw_read2count[t_seq], correct_read2count[t_seq]]
            #         for item in untouched_positive_seqs:
            #             correct_group.append([item, true_read2count[item], raw_read2count[item], correct_read2count[item]])  
            #         fn_lst.append([umi, str(true_group), str(raw_group), str(correct_group)])   

            #         generate_fp_seqs = untouched_positive_seqs - raw_positive_seqs   
            #         correct_group2 = []
            #         if len(generate_fp_seqs) > 0:
            #             for seq in generate_fp_seqs:
            #                 correct_group2.append([seq, true_read2count[seq], raw_read2count[seq], correct_read2count[seq]])  
            #             positive_fp_lst.append([umi, str(true_group), str(correct_group2)])                                       
            # # fp
            # elif len(raw_positive_seqs) == 0:
            #     if len(untouched_positive_seqs) > 0:
            #         true_group = [t_seq, true_read2count[t_seq], raw_read2count[t_seq], correct_read2count[t_seq]]
            #         correct_group = []
            #         for item in untouched_positive_seqs:
            #             correct_group.append([item, true_read2count[item], raw_read2count[item], correct_read2count[item]])  
            #         positive_fp_lst.append([umi, str(true_group), str(correct_group)])                    


            # if len(untouched_positive_seqs) > 0:
            #     intersec_seqs = untouched_positive_seqs & raw_positive_seqs
            #     true_group = [true_seq, true_read2count[true_seq], raw_read2count[true_seq], correct_read2count[true_seq]]
            #     raw_group = []
            #     if len(intersec_seqs) > 0:
            #         for item in intersec_seqs:
            #             raw_group.append([item, true_read2count[item], raw_read2count[item], correct_read2count[item]])
            #     correct_group = []
            #     for item in untouched_positive_seqs:
            #         correct_group.append([item, true_read2count[item], raw_read2count[item], correct_read2count[item]])
            #     fn_lst.append([umi, str(true_group), str(raw_group), str(correct_group)])
        #########
        # high_ambi_worksheet = self.workbook_flie.add_worksheet('high_ambiguous')
        # for r_n, data in enumerate(original_high_ambiguous):
        #     high_ambi_worksheet.write_row(r_n, 0, data)

        # fn_worksheet = self.workbook_flie.add_worksheet('fn')
        # for row_num, data in enumerate(fn_lst):
        #     fn_worksheet.write_row(row_num, 0, data)

        # fp_worksheet = self.workbook_flie.add_worksheet('positive_fp')
        # for row_num, data in enumerate(positive_fp_lst):
        #     fp_worksheet.write_row(row_num, 0, data)

        self.evaluation_metircs(base_level, 'Base level')
        self.evaluation_metircs(read_level, 'Read level') 
        rawset_entropy = self.set_entropy(len(raw_err_seqs), len(raw_errfreee_seqs), total_reads_num)
        correctset_entropy = self.set_entropy(len(correct_err_seqs), len(correct_errfree_seqs), total_reads_num)
        self.save_entropy('Purity entropy', rawset_entropy, correctset_entropy)
        return

    def evaluation_without_groundtruth(self, raw_read2count, correct_read2count, total_reads_num):
        """
        Evaluating an error correction method using raw and corrected data (before and after) base on defined read counts entropy

        Args:
            raw_read2count (dict): A dictionary to save reads (keys) and its counts (values) for raw dataset
            correct_read2count (dict): A dictionary to save reads (keys) and its counts (values) for corrected dataset of raw dataset
            total_reads_num (int): the number of the total reads
        """
        # rawset_entropy_noTruth = self.read_counts_entropy(raw_read2count, total_reads_num)
        # correctset_entropy_noTruth = self.read_counts_entropy(correct_read2count, total_reads_num)
        # self.save_entropy('ReadCountEntropy', rawset_entropy_noTruth, correctset_entropy_noTruth) 

        # entropy = self.noise2signal_entropy(raw_read2count, correct_read2count, total_reads_num)
        entropy = self.noise2read_entropy(raw_read2count, correct_read2count, total_reads_num)
        # self.save_entropy('Entropy H', entropy[0], entropy[1]) 
        self.logger.info("{}: raw dataset entropy: {}, correct dataset entropy: {}".format('Entropy H', entropy[0], entropy[1]))
        self.logger.info("Information Gain: {}".format(entropy[0] - entropy[1]))
        
        worksheet3 = self.workbook_flie.add_worksheet('Non-frequent Entropy')
        worksheet3.write('A1', 'H')
        worksheet3.write('A2', entropy[0])
        worksheet3.write('B1', "H'")
        worksheet3.write('B2', entropy[1]) 
        worksheet3.write('C1', '\u0394 H')
        worksheet3.write('C2', entropy[0] - entropy[1]) 

        self.workbook_flie.close()
        # total_variation = self.total_variation_differnce(raw_read2count, correct_read2count, total_reads_num)
        # print("Total variation distance: {}".format(total_variation))
        # relative_entropyq2p, entropyp2q = self.total_variation_differnce(raw_read2count, correct_read2count, total_reads_num)
        # entropy = self.spectral_entropy(raw_read2count, correct_read2count, total_reads_num)
        # raw_entropy = self.spectral_entropy(raw_read2count, raw_read2count, total_reads_num)
        
        return
        
    def evaluation_metircs(self, confusion_dict, sheet_name):
        """
        Evaluation metrics

        Args:
            confusion_dict (dict): A dictionary consists the values of TP, TN, FP and FN
            sheet_name (str): A string to indicate a sheet name of a csv file
        """
        tp = confusion_dict['tp']
        tn = confusion_dict['tn']
        fp = confusion_dict['fp']
        fn = confusion_dict['fn']
        if tp + tn + fp + fn != 0: 
            accuracy = (tp + tn) / (tp + tn + fp + fn)
        else:
            accuracy = "None"
        if tp + fp != 0:
            precision = tp / (tp + fp)
        else:
            precision = "None"
        if tp + fn != 0:
            recall = tp / (tp + fn)
            gain = (tp - fp) / (tp + fn)
        else:
            recall = "None"
            gain = "None"
        if fp + tn != 0:
            fall_out = fp / (fp + tn)
        else:
            fall_out = "None"
        # self.logger.info(sheet_name)
        self.logger.info("{}: accuracy: {}, precision: {}, recall: {}, gain: {}, fall-out: {}".format(sheet_name, accuracy, precision, recall, gain, fall_out))
        worksheet3 = self.workbook_flie.add_worksheet(sheet_name)
        worksheet3.write('A1', 'TP')
        worksheet3.write('A2', tp)
        worksheet3.write('B1', 'FP')
        worksheet3.write('B2', fp)
        worksheet3.write('C1', 'FN')
        worksheet3.write('C2', fn)
        worksheet3.write('D1', 'TN')
        worksheet3.write('D2', tn)

        worksheet3.write('E1', 'Accuracy')
        worksheet3.write('E2', accuracy)
        worksheet3.write('F1', 'Precision')
        worksheet3.write('F2', precision)
        worksheet3.write('G1', 'Recall')
        worksheet3.write('G2', recall)
        worksheet3.write('H1', 'Gain')
        worksheet3.write('H2', gain)
        worksheet3.write('I1', 'Fall-out')
        worksheet3.write('I2', fall_out)
        return

    def entropy(self, seqs_num, total_num):
        """
        part of the dataset entropy to measure the reads impurity of error-contained and error-free
        Args:
            seqs_num (int): the number of error-contained or error-free reads
            total_num (int): the number of total reads

        Returns:
            float: part of the dataset entropy
        """
        p = seqs_num / total_num
        if p == 1 or p == 0:
            entropy_val = 0
        else:
            entropy_val = -(p * math.log2(p))
        return entropy_val

    def gain2heatmap(self, correct2raw_diff):
        
        x_len = len(correct2raw_diff)
        x_len_sqrt = round(math.sqrt(x_len))
        xx = x_len_sqrt * x_len_sqrt
        val = x_len - xx
        if xx == x_len:
            m = x_len_sqrt
            n = x_len_sqrt
            x = np.array((correct2raw_diff))
            x_res=x.reshape(m, n)
 
        elif val < x_len_sqrt:
            m = x_len_sqrt
            n = x_len_sqrt + 1
            remain = (m - val) * [0]
            x = np.array((correct2raw_diff + remain))
            x_res = x.reshape(m, n)

        elif val > x_len_sqrt:
            m = x_len_sqrt
            n = x_len_sqrt + 2
            remain = (val - m) * [0]
            x = np.array((correct2raw_diff + remain))
            x_res=x.reshape(m, n)

        # thre_max = 0
        # thre_max = 1000000
        fig, ax = plt.subplots(figsize=(6, 6))
        # cmap = mpl.colors.ListedColormap(['whitesmoke', 'cornflowerblue', 'blueviolet', 'fuchsia', 'yellow', 'lawngreen', 'orange', 'darkorange', 'black', 'forestgreen', 'deepskyblue', 'yellow', 'orange'])
        # cmap = mpl.colors.ListedColormap([ '#8dd3c7', '#ffffb3', '#bebada', '#8dd3c7', '#ffffb3', '#bebada', '#fb8072', '#80b1d3', '#fdb462', '#b3de69', '#fccde5', 'deepskyblue', '#bc80bd', '#ccebc5', '#ffed6f', 'whitesmoke'])
        # name_lst = []
        # i = 0
        # for name, _ in mpl.colors.cnames.items():
        #     if name not in ['black', 'maroon', 'darked', 'darkgreen', 'darkslategray', 'darkslategrey', 'midnightblue', 'navy', 'darkblue', 'mediumblue']:
        #         name_lst.append(name)
        #     if i == 149:
        #         break
        #     i += 1
        # print(name_lst)
        # name_lst = ['aliceblue', 'antiquewhite', 'aqua', 'aquamarine', 'azure', 'beige', 'bisque', 'blanchedalmond', 'blue', 'blueviolet', 'brown', 'burlywood', 'cadetblue', 'chartreuse', 'chocolate', 'coral', 'cornflowerblue', 'cornsilk', 'crimson', 'cyan', 'deeppink', 'deepskyblue', 'dimgray', 'dimgrey', 'dodgerblue', 'firebrick', 'floralwhite', 'forestgreen', 'fuchsia', 'gainsboro', 'ghostwhite', 'gold', 'goldenrod', 'gray', 'green', 'greenyellow', 'grey', 'honeydew', 'hotpink', 'indianred', 'indigo', 'ivory', 'khaki', 'lavender', 'lavenderblush', 'lawngreen', 'lemonchiffon', 'lightblue', 'lightcoral', 'lightcyan', 'lightgoldenrodyellow', 'lightgray', 'lightgreen', 'lightgrey', 'lightpink', 'lightsalmon', 'lightseagreen', 'lightskyblue', 'lightslategray', 'lightslategrey', 'lightsteelblue', 'lightyellow', 'lime', 'limegreen', 'linen', 'magenta', 'mediumaquamarine', 'mediumorchid', 'mediumpurple', 'mediumseagreen', 'mediumslateblue', 'mediumspringgreen', 'mediumturquoise', 'mediumvioletred', 'mintcream', 'mistyrose', 'moccasin', 'navajowhite', 'oldlace', 'olive', 'olivedrab', 'orange', 'orangered', 'orchid', 'palegoldenrod', 'palegreen', 'paleturquoise', 'palevioletred', 'papayawhip', 'peachpuff', 'peru', 'pink', 'plum', 'powderblue', 'purple', 'rebeccapurple', 'red', 'rosybrown', 'royalblue', 'saddlebrown', 'salmon', 'sandybrown', 'seagreen', 'seashell', 'sienna', 'silver', 'skyblue', 'slateblue', 'slategray', 'slategrey',  'whitesmoke', '#ffffb3', '#bebada', '#fb8072', '#80b1d3', '#fdb462', '#b3de69', '#fccde5', '#bc80bd', '#ccebc5', '#ffed6f']
        # cmap = ListedColormap(name_lst)
        # cmap = mpl.colors.ListedColormap(['whitesmoke', '#8dd3c7', '#ffffb3', '#bebada', '#fb8072', '#80b1d3'])                                  
        # cmap.set_over('yellow')
        # cmap.set_under('red')
        # masked_array = np.ma.array (x_res, mask=np.isnan(x_res))
        # cur_cmap = mpl.cm.get_cmap(cmap)
        cur_cmap = mpl.cm.get_cmap('tab20c')#
        # cur_cmap.set_over('red')
        cur_cmap.set_under('red')
        cur_cmap.set_bad(color='red')
        image = ax.imshow(x_res, cmap=cur_cmap) #interpolation='none', , vmax=thre_max
        # image = ax.pcolor(x_res, cmap=cmap, vmin=thre_min, antialiased=True)
        cbar = fig.colorbar(image, extend='min', shrink=0.8)
        cbar.cmap.set_under('red')
        # cbar.cmap.set_over('red')
        fig.tight_layout()
        fig.savefig(os.path.join(self.config.result_dir, self.prefix + '_information_gain.png'), transparent=True)

        return

    def entropy_item(self, total_freq, freq):
        if freq > 0:
            p = freq / total_freq
            result = - p * math.log2(p)
        return result

    def noise2read_entropy(self, raw_read2count, correct_read2count, total_num):

        raw_unique_reads = set(raw_read2count.keys())
        correct_unique_reads = set(correct_read2count.keys())

        frequent_reads = set([k for k, v in raw_read2count.items() if v > self.config.high_freq_thre])

        # raw dataset
        non_frequent_raw_reads = raw_unique_reads - frequent_reads
        raw_entropy_items = []
        for read in non_frequent_raw_reads:
            raw_entropy_items.append(raw_read2count[read])

        raw_nonFre_reads_total_num = sum(raw_entropy_items)
        # raw entropy
        with WorkerPool(self.config.num_workers, shared_objects=raw_nonFre_reads_total_num, start_method='fork') as pool:
            raw_entropy_lst = pool.map(self.entropy_item, raw_entropy_items)
        raw_entropy = sum(raw_entropy_lst) 

        # correct dateset
        non_frequent_correct_reads = correct_unique_reads - frequent_reads
        correct_entropy_items = []
        for read in non_frequent_correct_reads:
            correct_entropy_items.append(correct_read2count[read])

        # correct entropy
        correct_nonFre_reads_total_num = sum(correct_entropy_items)

        with WorkerPool(self.config.num_workers, shared_objects=correct_nonFre_reads_total_num, start_method='fork') as pool:
            correct_entropy_lst = pool.map(self.entropy_item, correct_entropy_items) 
        correct_entropy = sum(correct_entropy_lst)
        ##################################################################################
        #information gain (\delta I) heatmap
        new_reads = correct_unique_reads - raw_unique_reads
        new_reads_num = len(new_reads)
        self.logger.info("Wrongly introduced {} new reads".format(new_reads_num))

        raw_kept_counts = []
        correct_kept_counts = []
        kept_reads = correct_unique_reads & raw_unique_reads
        for read in kept_reads:
            correct_kept_counts.append(correct_read2count[read])
            raw_kept_counts.append(raw_read2count[read])

        with WorkerPool(self.config.num_workers, shared_objects=total_num, start_method='fork') as pool:
            raw_kept_entropy_lst = pool.map(self.entropy_item, raw_kept_counts)

        with WorkerPool(self.config.num_workers, shared_objects=total_num, start_method='fork') as pool:
            correct_kept_entropy_lst = pool.map(self.entropy_item, correct_kept_counts) 

        raw_removed_reads = raw_unique_reads - correct_unique_reads
        raw_removed_items = []
        for read in raw_removed_reads:
            raw_removed_items.append(raw_read2count[read])

        with WorkerPool(self.config.num_workers, shared_objects=total_num, start_method='fork') as pool:
            raw_removed_entropy_items_lst = pool.map(self.entropy_item, raw_removed_items)
        for i in raw_removed_entropy_items_lst:
            if i <=0:
                print('Warning')
        entropy_item_lst = []
        for i, j in zip(raw_kept_entropy_lst, correct_kept_entropy_lst):
            entropy_item_lst.append(i - j)
        entropy_item_lst.extend(raw_removed_entropy_items_lst)
        if new_reads_num > 0:
            entropy_item_lst.extend([np.nan] * new_reads_num)
        shuffle(entropy_item_lst)
        self.gain2heatmap(entropy_item_lst)
        return [raw_entropy, correct_entropy]

    def set_entropy(self, err_seqs_num, errfree_seqs_num, total_num):
        """
        dataset entropy to measure the reads impurity of error-contained and error-free

        Args:
            err_seqs_num (int): the number of error-conatined reads
            errfree_seqs_num (int): the number of error-free reads
            total_num (int): the number of total reads

        Returns:
            float: dataset entropy
        """
        # total_seqs = err_seqs + errfree_seqs
        # unique_len = len(set(total_seqs))
        err_entropy = self.entropy(err_seqs_num, total_num)
        errfree_entropy = self.entropy(errfree_seqs_num, total_num)
        return errfree_entropy + err_entropy

    def save_entropy(self, sheet_name, raw_set_entropy, correct_entropy):
        """
        save entropy result to a csv sheet

        Args:
            sheet_name (str): A string to indicate a sheet name of a csv file
            raw_set_entropy (float): raw dataset entropy
            correct_entropy (float): corrected dataset entropy
        """
        self.logger.info("{}: raw dataset entropy: {}, correct dataset entropy: {}".format(sheet_name, raw_set_entropy, correct_entropy))
        worksheet3 = self.workbook_flie.add_worksheet(sheet_name)
        worksheet3.write('A1', 'E')
        worksheet3.write('A2', raw_set_entropy)
        worksheet3.write('B1', "E'")
        worksheet3.write('B2', correct_entropy) 
        worksheet3.write('C1', '\u0394 E')
        worksheet3.write('C2', raw_set_entropy - correct_entropy) 
        return   

    #############################################################################################################
    # Warning: the following functions have been deprecated and may contain bugs. If you want to use, you must check every row carefully.

    '''
    def noise2signal_entropy(self, raw_read2count, correct_read2count, total_num):

        raw_unique_reads = set(list(raw_read2count.keys()))
        correct_unique_reads = set(list(correct_read2count.keys()))
        total_unique_reads = raw_unique_reads.union(correct_unique_reads)
        id2read = {}
        idx = 0
        for read in total_unique_reads:
            id2read[idx] = read
            idx += 1
    
        raw_only_reads = raw_unique_reads - correct_unique_reads
        correct_only_reads = correct_unique_reads - raw_unique_reads
        raw_correct_inter = correct_unique_reads & raw_unique_reads

        for i in len(total_unique_reads):
            read = id2read[i]
            if read in 

        entropy = 0
        n = len(raw_unique_reads)
        correct_in_counts = []
        out_counts = []        
        for read in correct_unique_reads:
            correct_count = correct_read2count[read]
            # p = correct_count / total_num  
            if read in raw_unique_reads:
                correct_in_counts.append(correct_count)
            else:
                out_counts.append(correct_count)

        correct_in_counts.sort(reverse=True)
        out_counts.sort(reverse=True)

        shared_vars = total_num, n, 1
        with WorkerPool(self.config.num_workers, shared_objects=shared_vars, start_method='fork') as pool:
            in_entropy = pool.map(self.entropy_item, in_counts)
        if len(out_counts) > 0:
            shared_vars = total_num, n, self.config.delta
            with WorkerPool(self.config.num_workers, shared_objects=shared_vars, start_method='fork') as pool:
                out_entropy = pool.map(self.entropy_item, out_counts)
            entropy = sum(in_entropy) + sum(out_entropy)
            shared_vars = total_num, n, 1
            with WorkerPool(self.config.num_workers, shared_objects=shared_vars, start_method='fork') as pool:
                out_entropy2 = pool.map(self.entropy_item, out_counts)
        else:
            entropy = sum(in_entropy)

        # for read in correct_unique_reads:
        #     correct_count = correct_read2count[read]
        #     p = correct_count / total_num  
        #     if read in raw_unique_reads:
        #         entropy += -(p * math.log2(p) / math.log2(n))
        #     else:
        #         entropy += -10*(p * math.log2(p) / math.log2(n))
        # print(entropy)
        # visualisation
        if raw_read2count == correct_read2count:
            if len(out_counts) > 0:
                raw_lst = in_entropy + len(out_entropy2) * [0]
            else:
                raw_lst = in_entropy
            raw_heatmap = os.path.join(self.config.result_dir, 'raw_entropy_heatmap.png')
            self.entropy_visualisation(raw_lst, )
        else:
            correct_lst = in_entropy + out_entropy2
            correct_heatmap = os.path.join(self.config.result_dir, 'correct_entropy_heatmap.png')
            self.entropy_visualisation(correct_lst, correct_heatmap)
        return entropy
    '''

    '''
    def evaluation_with_groundtruth0(self):
        read_level = {'tp':0, 'fp':0, 'tn':0, 'fn':0}
        base_level = {'tp':0, 'fp':0, 'tn':0, 'fn':0}
        true_records = SeqIO.index(self.config.ground_truth_data, parse_file_type(self.config.ground_truth_data))
        error_records = SeqIO.index(self.config.input_file, parse_file_type(self.config.input_file))
        correct_records = SeqIO.index(self.config.correct_data, parse_file_type(self.config.input_file))
        total_reads_num = 0
        # for calculating entropy
        correct_errfree_seqs = []
        correct_err_seqs = []
        raw_errfreee_seqs = []
        raw_err_seqs = []

        true_seqs_lst = []
        raw_seqs_lst = []
        correct_seqs_lst = []
        for name in error_records:
            true_seq = str(true_records[name].seq)
            raw_seq = str(error_records[name].seq)
            correct_seq = str(correct_records[name].seq)   
            true_seqs_lst.append(true_seq)     
            raw_seqs_lst.append(raw_seq)
            correct_seqs_lst.append(correct_seq)
        raw_read2count = collections.Counter(raw_seqs_lst)
        true_read2count = collections.Counter(true_seqs_lst)
        correct_read2count = collections.Counter(correct_seqs_lst)

        fn_lst = []

        for name in error_records:
            total_reads_num += 1
            true_seq = str(true_records[name].seq)
            raw_seq = str(error_records[name].seq)
            correct_seq = str(correct_records[name].seq)

            true_raw_dis = editdistance.eval(true_seq, raw_seq)
            true_cor_dis = editdistance.eval(true_seq, correct_seq)
            raw_cor_dis = editdistance.eval(raw_seq, correct_seq)
            #
            if true_cor_dis == 0:
                correct_errfree_seqs.append(correct_seq)
            else:
                correct_err_seqs.append(correct_seq)

            # actual condition with no errors
            # negative conditions
            if true_raw_dis == 0:
                raw_errfreee_seqs.append(raw_seq)
                if true_cor_dis == 0:
                    base_level['tn'] += len(true_seq)
                    read_level['tn'] += 1
                else:
                    base_level['tn'] += len(true_seq) - true_cor_dis
                    base_level['fp'] += true_cor_dis
                    read_level['fp'] += 1
            # true_raw_dis != 0 which means actual condition contain errors 
            else:
                raw_err_seqs.append(raw_seq)
                if true_cor_dis == 0:
                    base_level['tp'] += true_raw_dis
                    base_level['tn'] += len(true_seq) - true_raw_dis
                    read_level['tp'] += 1
                else:
                    read_level['fn'] += 1
                    fn_lst.append([raw_seq, raw_read2count[raw_seq], true_seq, true_read2count[true_seq], correct_seq, correct_read2count[correct_seq]])
                    if raw_cor_dis == 0:
                        base_level['tn'] += len(true_seq) - true_raw_dis
                        base_level['fn'] += true_raw_dis # true_raw_dis == true_cor_dis
                    else:
                        # measure the distance change with ture sequence before and after correction
                        gap = true_raw_dis - true_cor_dis
                        if gap > 0: # distance decrease after correction
                            base_level['tp'] += gap
                            base_level['tn'] += len(true_seq) - true_raw_dis
                            base_level['fn'] += true_cor_dis
                        elif gap < 0:
                            base_level['tn'] += len(true_seq) - true_cor_dis
                            base_level['fp'] += abs(gap)
                            base_level['fn'] += true_raw_dis
                        else:
                            base_level['tn'] += len(true_seq) - true_cor_dis
                            base_level['fn'] += true_cor_dis
        worksheet = self.workbook_flie.add_worksheet('fn')
        for row_num, data in enumerate(fn_lst):
            worksheet.write_row(row_num, 0, data)
        self.evaluation_metircs(base_level, 'Base level')
        self.evaluation_metircs(read_level, 'Read level') 
        rawset_entropy = self.set_entropy(raw_err_seqs, raw_errfreee_seqs, total_reads_num)
        correctset_entropy = self.set_entropy(correct_err_seqs, correct_errfree_seqs, total_reads_num)
        self.save_entropy('Dataset Entropy', rawset_entropy, correctset_entropy)
        # self.workbook_flie.close()

        return
    '''        
    '''
    # def entropy_visualisation(self, x_lst, out_heatmap, v_min, v_max):
    def entropy_visualisation(self, raw_lst, correct_lst):
        val_unique = list(set(raw_lst + correct_lst))

        norm_raw_lst = []
        for raw_v in raw_lst:
            if raw_v == 0:
                cur_bin_num = 0
            else:
                inter_val = int(raw_v / 5)
                cur_bin_num = 0
                for i in range(inter_val):
                    cur_bin_num += i+1
                    if cur_bin_num > inter_val:
                        cur_bin_num -= 1
                        break
            norm_raw_lst.append(cur_bin_num)

        norm_correct_lst = []
        for correct_v in correct_lst:
            if correct_v == 0:
                cur_bin_num = 0
            else:
                inter_val = int(correct_v / 5)
                cur_bin_num = 0
                for i in range(inter_val):
                    cur_bin_num += i+1
                    if cur_bin_num > inter_val:
                        cur_bin_num -= 1
                        break
            norm_correct_lst.append(cur_bin_num)

        # color_num = len(norm_raw_lst)
        
        x_len = len(norm_raw_lst)
        x_len_sqrt = round(math.sqrt(x_len))
        xx = x_len_sqrt * x_len_sqrt
        val = x_len - xx
        if xx == x_len:
            m = x_len_sqrt
            n = x_len_sqrt
            raw_x = np.array((norm_raw_lst))
            raw_x_res=raw_x.reshape(m, n)

            correct_x = np.array((norm_correct_lst))
            correct_x_res=correct_x.reshape(m, n)
            
        elif val < x_len_sqrt:
            m = x_len_sqrt
            n = x_len_sqrt + 1
            remain = (m - val) * [0]
            raw_x = np.array((norm_raw_lst + remain))
            raw_x_res = raw_x.reshape(m, n)

            correct_x = np.array((norm_correct_lst + remain))
            correct_x_res = correct_x.reshape(m, n)

        elif val > x_len_sqrt:
            m = x_len_sqrt
            n = x_len_sqrt + 2
            remain = (val - m) * [0]
            raw_x = np.array((norm_raw_lst + remain))
            raw_x_res=raw_x.reshape(m, n)

            correct_x = np.array((norm_correct_lst + remain))
            correct_x_res=correct_x.reshape(m, n)

        # x_len = len(correct_lst)
        # x_len_sqrt = round(math.sqrt(x_len))
        # xx = x_len_sqrt * x_len_sqrt
        # val = x_len - xx
        # if xx == x_len:
        #     m = x_len_sqrt
        #     n = x_len_sqrt
        #     correct_x = np.array((correct_lst))
        #     correct_x_res=correct_x.reshape(m, n)
            
        # elif val < x_len_sqrt:
        #     m = x_len_sqrt
        #     n = x_len_sqrt + 1
        #     remain = (m - val) * [0]
        #     correct_x = np.array((correct_lst + remain))
        #     correct_x_res = correct_x.reshape(m, n)

        # elif val > x_len_sqrt:
        #     m = x_len_sqrt
        #     n = x_len_sqrt + 2
        #     remain = (val - m) * [0]
        #     correct_x = np.array((correct_lst + remain))
        #     correct_x_res=correct_x.reshape(m, n)


        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(32,16)) #

        fig.subplots_adjust(wspace=0.01)
        # unique_val = set(x_lst)
        # color_num = len(unique_val)
        # print(color_num)
        # cmap = sns.color_palette(n_colors = color_num, linecolor='white')#"deep",
        # cmap = sns.color_palette("Paired")
        # cmap = LinearSegmentedColormap.from_list(N=256)

        # color_num = len(set(norm_raw_lst))
        max_v1 = max(norm_raw_lst)
        max_v2 = max(norm_correct_lst)
        color_num = max(max_v1, max_v2) + 1
        name_lst = []
        i = 0
        for name, _ in mpl.colors.cnames.items():
            name_lst.append(name)
            if i == color_num-1:
                break
        cmap = ListedColormap(name_lst)

        v_min = min(val_unique)
        v_max = max(val_unique)
        sns.heatmap(raw_x_res, square=True, ax=ax1, vmin=v_min, vmax=v_max)
        # sns.heatmap(raw_x_res, square=True, ax=ax1, cmap=cmap, cbar=True)
        # fig.colorbar(ax1.collections[0], ax=ax1, location="left", use_gridspec=False, pad=0.2) #, pad=0.2 ticks=val_unique, 

        # color_num2 = len(set(correct_lst))
        # name_lst2 = []
        # i = 0
        # for name2, _ in mpl.colors.cnames.items():
        #     name_lst2.append(name2)
        #     if i == color_num2-1:
        #         break
        # cmap2 = ListedColormap(name_lst2)
        # sns.heatmap(correct_x_res, square=True, cmap=cmap, ax=ax2, cbar=True)
        sns.heatmap(correct_x_res, square=True, ax=ax2, vmin=v_min, vmax=v_max)
        # fig.colorbar(ax2.collections[0], ax=ax2, location="right", use_gridspec=False, pad=0.2) #, pad=0.2  ticks=val_unique,
        # colorbar = ax1.collections[0].colorbar
        # colorbar.set_ticks(val_unique)
        # colorbar.set_ticklabels(val_unique)

        # plt.tight_layout()
        plt.savefig(os.path.join(self.config.result_dir, 'entropy_heatmap.png')) 
        plt.close()
        return
    def noise2signal_entropy0(self, raw_read2count, correct_read2count, total_num):

        raw_unique_reads = list(raw_read2count.keys())
        correct_unique_reads = list(correct_read2count.keys())

        new_reads = list(set(correct_unique_reads) - set(raw_unique_reads))
        raw_unique_reads.sort()
        new_reads.sort()

        raw_in_counts = []
        correct_in_counts = []
        correct_out_counts = []
        for read in raw_unique_reads:
            raw_in_counts.append(raw_read2count[read])
            if read in correct_unique_reads:
                correct_in_counts.append(correct_read2count[read])
            else:
                correct_in_counts.append(0) # noise of identical reads removed by marked as 0
        if len(new_reads) > 0:
            for read in new_reads:
                correct_out_counts.append(correct_read2count[read])
            raw_out_counts = [0] * len(new_reads) # noise of new sequence generated by marked as 0

        # raw entropy
        n = len(raw_unique_reads)
        shared_vars = total_num, n, 1
        with WorkerPool(self.config.num_workers, shared_objects=shared_vars, start_method='fork') as pool:
            raw_entropy_lst = pool.map(self.entropy_item, raw_in_counts)
        # correct entropy
        with WorkerPool(self.config.num_workers, shared_objects=shared_vars, start_method='fork') as pool:
            correct_in_entropy_lst = pool.map(self.entropy_item, correct_in_counts) 

        raw_entropy = sum(raw_entropy_lst)
        new_seq_len = len(correct_out_counts) 
        if new_seq_len > 0:
            shared_vars = total_num, n, self.config.delta
            with WorkerPool(self.config.num_workers, shared_objects=shared_vars, start_method='fork') as pool:
                correct_out_entropy_lst = pool.map(self.entropy_item, correct_out_counts)

            correct_entropy = sum(correct_out_entropy_lst)
            for v in correct_in_entropy_lst:
                if v != -2:
                    correct_entropy += v
        else:
            correct_entropy = 0
            for v in correct_in_entropy_lst:
                if v != -2:
                    correct_entropy += v

        if new_seq_len > 0:
            # raw
            # raw_lst = copy.deepcopy(raw_entropy_lst + raw_out_counts)
            raw_lst = raw_in_counts + raw_out_counts
            # raw_lst = raw_in_counts
            # correct_lst = copy.deepcopy(correct_in_entropy_lst + correct_out_entropy_lst)
            correct_lst = correct_in_counts + correct_out_counts
        else:
            # raw_lst = copy.deepcopy(raw_entropy_lst)
            # correct_lst = copy.deepcopy(correct_in_entropy_lst)
            raw_lst = raw_in_counts
            correct_lst = correct_in_counts   
        self.entropy_visualisation(raw_lst, correct_lst)
        # raw_heatmap = os.path.join(self.config.result_dir, 'raw_entropy_heatmap.png')

        # self.entropy_visualisation(raw_lst, raw_heatmap, len(set(raw_lst)))
        # # correct
        # correct_heatmap = os.path.join(self.config.result_dir, 'correct_entropy_heatmap.png')
        # # self.entropy_visualisation(correct_lst, correct_heatmap,v_min, v_max)
        # # self.entropy_visualisation(correct_lst, correct_heatmap, color_num)
        # self.entropy_visualisation(correct_lst, correct_heatmap, len(set(correct_lst)))
        return [raw_entropy, correct_entropy]
    '''
    '''
    def noise2read_entropy(self, raw_read2count, correct_read2count):

        raw_unique_reads = set(raw_read2count.keys())
        correct_unique_reads = set(correct_read2count.keys())

        frequent_reads = set([k for k, v in raw_read2count.items() if v > self.config.high_freq_thre])

        # raw dataset
        non_frequent_raw_reads = raw_unique_reads - frequent_reads
        raw_entropy_items = []
        
        for read in non_frequent_raw_reads:
            raw_entropy_items.append(raw_read2count[read])

        raw_nonFre_reads_num = sum(raw_entropy_items)
        # raw entropy
        n = len(non_frequent_raw_reads)
        shared_vars1 = raw_nonFre_reads_num, n, 1
        with WorkerPool(self.config.num_workers, shared_objects=shared_vars1, start_method='fork') as pool:
            raw_entropy_lst = pool.map(self.entropy_item, raw_entropy_items)
        raw_entropy = sum(raw_entropy_lst) 

        # correct dateset
        non_frequent_correct_reads = correct_unique_reads - frequent_reads

        correct_entropy_items = []

        for read in non_frequent_correct_reads:
            correct_entropy_items.append(correct_read2count[read])

        # correct entropy
        correct_nonFre_reads_num = len(non_frequent_correct_reads)
        shared_vars2 = correct_nonFre_reads_num, n, 1

        with WorkerPool(self.config.num_workers, shared_objects=shared_vars2, start_method='fork') as pool:
            correct_entropy_lst = pool.map(self.entropy_item, correct_entropy_items) 
        correct_entropy = sum(correct_entropy_lst)
        ##################################################################################
        new_reads = correct_unique_reads - raw_unique_reads
        new_reads_num = len(new_reads)
        self.logger.info("Wrongly introduced {} new reads".format(new_reads_num))

        # information gain visualisation
        correct2raw_diff = []
        common_non_frequent_reads = non_frequent_raw_reads & non_frequent_correct_reads

        common_non_frequent_raw_items = []
        common_non_frequent_correct_items = []
        for read in common_non_frequent_reads:
            common_non_frequent_raw_items.append(raw_read2count[read])
            common_non_frequent_correct_items.append(correct_read2count[read])

        with WorkerPool(self.config.num_workers, shared_objects=shared_vars1, start_method='fork') as pool:
            raw_common_lst = pool.map(self.entropy_item, common_non_frequent_raw_items) 

        with WorkerPool(self.config.num_workers, shared_objects=shared_vars2, start_method='fork') as pool:
            correct_common_lst = pool.map(self.entropy_item, common_non_frequent_correct_items) 

        if new_reads_num > 0:
            correct2raw_diff.extend([np.nan] * new_reads_num)

        ###### removed reads
        # removed_reads = non_frequent_raw_reads - non_frequent_correct_reads
        # if len(removed_reads) > 0:
        #     correct2raw_diff.extend([1000000] * len(removed_reads))

        for i, j in zip(raw_common_lst, correct_common_lst):
            correct2raw_diff.append(i - j)

        shuffle(correct2raw_diff)
        self.gain2heatmap(correct2raw_diff)

        return [raw_entropy, correct_entropy]
    '''

    '''
    def fft_spectrum(self, freq):
        return freq * freq

    def spectral_entropy(self, raw_read2count, correct_read2count, total_num):

        raw_unique_reads = list(raw_read2count.keys())
        correct_unique_reads = list(correct_read2count.keys())
        n = len(raw_unique_reads)

        in_counts = []
        out_counts = []
        counts = []
        for read in correct_unique_reads:
            correct_count = correct_read2count[read]
            p = correct_count / total_num  
            if read in raw_unique_reads:
                in_counts.append(correct_count)
            else:
                out_counts.append(correct_count)
            counts.append(correct_count)
        # print(ant.spectral_entropy(in_counts, sf=100, method='welch', normalize=True))
        # print(ant.spectral_entropy(out_counts, sf=100, method='welch', normalize=True))
        # print(ant.spectral_entropy(counts, sf=100, method='welch', normalize=True))
        with WorkerPool(self.config.num_workers) as pool:
            in_fft_vals = pool.map(self.fft_spectrum, rfft(in_counts).tolist())
        if len(out_counts) > 0:
            with WorkerPool(self.config.num_workers) as pool:
                out_fft_vals = pool.map(self.fft_spectrum, rfft(out_counts).tolist())

            val_sum = sum(in_fft_vals) + sum(out_fft_vals)
        else:
            val_sum = sum(in_fft_vals)

        shared_vars = val_sum, n, 1
        with WorkerPool(self.config.num_workers, shared_objects=shared_vars, start_method='fork') as pool:
            in_entropy = pool.map(self.spectrum2entropy, in_fft_vals)
        if len(out_counts) > 0:
            shared_vars = val_sum, n, 10
            with WorkerPool(self.config.num_workers, shared_objects=shared_vars, start_method='fork') as pool:
                out_entropy = pool.map(self.spectrum2entropy, out_fft_vals)
            entropy = sum(in_entropy) + sum(out_entropy)
        else:
            entropy = sum(in_entropy)
        
        print(entropy)
        # return 

    def spectrum2entropy(self, shared_vars, freq):
        total_freq, n, delta = shared_vars
        p = freq / total_freq
        result = - delta * (p * cmath.log(p,2)/ cmath.log(n,2))
        return result

    def gain2heatmap0(self, correct2raw_diff):
        
        x_len = len(correct2raw_diff)
        x_len_sqrt = round(math.sqrt(x_len))
        xx = x_len_sqrt * x_len_sqrt
        val = x_len - xx
        if xx == x_len:
            m = x_len_sqrt
            n = x_len_sqrt
            x = np.array((correct2raw_diff))
            x_res=x.reshape(m, n)
 
        elif val < x_len_sqrt:
            m = x_len_sqrt
            n = x_len_sqrt + 1
            remain = (m - val) * [0]
            x = np.array((correct2raw_diff + remain))
            x_res = x.reshape(m, n)

        elif val > x_len_sqrt:
            m = x_len_sqrt
            n = x_len_sqrt + 2
            remain = (val - m) * [0]
            x = np.array((correct2raw_diff + remain))
            x_res=x.reshape(m, n)

        # Specify the indices of the two types of bad data
        outliers1 = np.where(x_res == -1000000)
        # outliers2 = np.where(x_res == 1000000)

        # Plot the heatmap
        sns.heatmap(x_res, cmap="Set3")

        # Plot the first type of outliers in red
        plt.scatter(outliers1[0], outliers1[1], color="red", s=200)

        # Plot the second type of outliers in blue
        # plt.scatter(outliers2[0], outliers2[1], color="blue", s=200)
        plt.savefig(os.path.join(self.config.result_dir, self.prefix + '_information_gain.png'), transparent=True)

        # # Plot the first type of bad data using a red color
        # fig, ax = plt.subplots(figsize=(6, 6))
        # x_res[bad_indices_1] = np.nan

        # cur_cmap = mpl.cm.get_cmap('Set3')
        # cur_cmap.set_under('red')
        # cur_cmap.set_bad(color='red')
        # heatmap1 = ax.imshow(x_res, cmap=cur_cmap)

        # # Plot the second type of bad data using a green color
        # x_res[bad_indices_2] = np.nan
        # cur_cmap = mpl.cm.get_cmap('Set3')
        # cur_cmap.set_over('black')
        # cur_cmap.set_bad(color='Black')
        # heatmap2 = ax.imshow(x_res, cmap=cur_cmap)

        # # image = ax.imshow(x_res, cmap=cur_cmap) #interpolation='none', , vmax=thre_max
        # # # image = ax.pcolor(x_res, cmap=cmap, vmin=thre_min, antialiased=True)

        # cbar = fig.colorbar(heatmap1, extend='both', shrink=0.8)
        # cbar.cmap.set_under('red')
        # cbar.cmap.set_over('black')
        # fig.tight_layout()
        # fig.savefig(os.path.join(self.config.result_dir, self.prefix + '_information_gain.png'), transparent=True)

        return

    '''
    '''
    def total_variation_differnce(self, raw_read2count, correct_read2count, total_num):

        raw_unique_reads = list(raw_read2count.keys())
        correct_unique_reads = list(correct_read2count.keys())
        total_unique_reads = set(raw_unique_reads + correct_unique_reads)
        variation = 0
        dif_vals = []
        for read in total_unique_reads:
            if read in raw_unique_reads:
                raw_count = raw_read2count[read]
            else:
                raw_count = 0
            if read in correct_unique_reads:
                correct_count = correct_read2count[read]
            else:
                correct_count = 0
            q = raw_count / total_num
            p = correct_count / total_num   
            dif_vals.append(p-q) 
        return norm(array(dif_vals))

    def relative_entropy(self, raw_read2count, correct_read2count, total_num):

        raw_unique_reads = list(raw_read2count.keys())
        correct_unique_reads = list(correct_read2count.keys())
        total_unique_reads = raw_unique_reads + correct_unique_reads
        relative_entropy = 0
        relative_entropy_p2q = 0
        for read in total_unique_reads:
            if read in raw_unique_reads:
                raw_count = raw_read2count[read]
            else:
                raw_count = 0
            if read in correct_unique_reads:
                correct_count = correct_read2count[read]
            else:
                correct_count = 0
            q = raw_count / total_num
            p = correct_count / total_num   
            if q != 0:
                relative_entropy += p*math.log2(p/q)   
            if p != 0: 
                relative_entropy_p2q += q*math.log2(q/p) 
        print(relative_entropy, relative_entropy_p2q)
        return 

    '''

    '''
    def read_counts_entropy(self, read_count, total_num):
        """
        read count entropy calculation

        Args:
            read_count (dict): A dictionary to save reads (keys) and its counts (values) for dataset
            total_num (int): the number of total reads

        Returns:
            float: read count entropy 
        """
        entropy_val = 0
        # n = len(read_count.values())
        for count in read_count.values():
            p = count / total_num
            entropy_val += -(p * math.log2(p))
            # entropy_val += -math.log2(p)
            # entropy_val += -(p * math.log2(p) / math.log2(n))
            # entropy_val += -(p * math.log10(p))
        return entropy_val

    def noise2signal_entropy(self, raw_read2count, correct_read2count, total_num):

        raw_unique_reads = list(raw_read2count.keys())
        correct_unique_reads = list(correct_read2count.keys())

        new_reads = list(set(correct_unique_reads) - set(raw_unique_reads))
        raw_unique_reads.sort()
        new_reads.sort()

        raw_counts = []
        raw_kept_counts = []
        correct_kept_counts = []
        
        raw_entropy_items = []

        correct2raw_diff = []

        correct_new_counts = []
        for read in raw_unique_reads:
            raw_counts.append(raw_read2count[read])
            if read in correct_unique_reads:
                correct_kept_counts.append(correct_read2count[read])
                raw_kept_counts.append(raw_read2count[read])
                correct2raw_diff.append( correct_read2count[read] - raw_read2count[read])
            else:
                # correct_removed_counts.append(0) # noise of identical reads removed by marked as 0
                # raw_removed_counts.append(raw_read2count[read])
                correct2raw_diff.append( 0 - raw_read2count[read])
                raw_entropy_items.append(raw_read2count[read])

        new_reads_num = len(new_reads)
        self.logger.info("Wrongly introduced {} new reads".format(new_reads_num))
        if new_reads_num > 0:
            entropy_new_items = []
            for read in new_reads:
                correct_new_counts.append(correct_read2count[read])
                correct2raw_diff.append(np.nan)
                entropy_new_items.append(np.nan)

        # raw entropy
        n = len(raw_unique_reads)
        shared_vars = total_num, n, 1
        with WorkerPool(self.config.num_workers, shared_objects=shared_vars, start_method='fork') as pool:
            raw_entropy_lst = pool.map(self.entropy_item, raw_counts)
        raw_entropy = sum(raw_entropy_lst) 

        with WorkerPool(self.config.num_workers, shared_objects=shared_vars, start_method='fork') as pool:
            raw_kept_entropy_lst = pool.map(self.entropy_item, raw_kept_counts)
        # correct entropy
        with WorkerPool(self.config.num_workers, shared_objects=shared_vars, start_method='fork') as pool:
            correct_kept_entropy_lst = pool.map(self.entropy_item, correct_kept_counts) 

        if new_reads_num > 0:
            shared_vars = total_num, n, self.config.delta
            with WorkerPool(self.config.num_workers, shared_objects=shared_vars, start_method='fork') as pool:
                correct_new_entropy_lst = pool.map(self.entropy_item, correct_new_counts)
            correct_entropy = sum(correct_new_entropy_lst) + sum(correct_kept_entropy_lst)
        else:
            correct_entropy = sum(correct_kept_entropy_lst)
            # correct_new_entropy_lst = []

        # shuffle(correct2raw_diff)
        # self.gain2heatmap(correct2raw_diff)
        with WorkerPool(self.config.num_workers, shared_objects=shared_vars, start_method='fork') as pool:
            raw_entropy_items_lst = pool.map(self.entropy_item, raw_entropy_items)
        for i in raw_entropy_items_lst:
            if i <=0:
                print('no')
        entropy_item_lst = []
        for i, j in zip(raw_kept_entropy_lst, correct_kept_entropy_lst):
            entropy_item_lst.append(i - j)
        entropy_item_lst.extend(raw_entropy_items_lst)
        if new_reads_num > 0:
            entropy_item_lst.extend(entropy_new_items)
        shuffle(entropy_item_lst)
        self.gain2heatmap(entropy_item_lst)

        return [raw_entropy, correct_entropy]

    def entropy_item(self, shared_vars, freq):
        total_freq, n, delta = shared_vars
        if freq > 0:
            p = freq / total_freq
            result = - delta * (p * math.log2(p)/ math.log2(n))
        return result


    def entropy_item(self, shared_vars, freq):
        total_freq, n = shared_vars
        if freq > 0:
            p = freq / total_freq
            # result = - p * math.log2(p) / math.log2(n)
            result = - p * math.log2(p)
        return result

    def noise2read_entropy(self, raw_read2count, correct_read2count, total_num):

        raw_unique_reads = set(raw_read2count.keys())
        correct_unique_reads = set(correct_read2count.keys())

        frequent_reads = set([k for k, v in raw_read2count.items() if v > self.config.high_freq_thre])

        # raw dataset
        non_frequent_raw_reads = raw_unique_reads - frequent_reads
        raw_entropy_items = []
        for read in non_frequent_raw_reads:
            raw_entropy_items.append(raw_read2count[read])

        raw_nonFre_reads_total_num = sum(raw_entropy_items)
        # raw entropy
        non_frequent_raw_unique_num = len(non_frequent_raw_reads)

        shared_vars1 = raw_nonFre_reads_total_num, non_frequent_raw_unique_num
        with WorkerPool(self.config.num_workers, shared_objects=shared_vars1, start_method='fork') as pool:
            raw_entropy_lst = pool.map(self.entropy_item, raw_entropy_items)
        raw_entropy = sum(raw_entropy_lst) 

        # correct dateset
        non_frequent_correct_reads = correct_unique_reads - frequent_reads
        correct_entropy_items = []
        for read in non_frequent_correct_reads:
            correct_entropy_items.append(correct_read2count[read])

        # correct entropy
        correct_nonFre_reads_total_num = sum(correct_entropy_items)

        # correct_unique_num = len(non_frequent_correct_reads)
        shared_vars2 = correct_nonFre_reads_total_num, non_frequent_raw_unique_num

        with WorkerPool(self.config.num_workers, shared_objects=shared_vars2, start_method='fork') as pool:
            correct_entropy_lst = pool.map(self.entropy_item, correct_entropy_items) 
        correct_entropy = sum(correct_entropy_lst)
        ##################################################################################
        #information gain (\delta I) heatmap
        new_reads = correct_unique_reads - raw_unique_reads
        new_reads_num = len(new_reads)
        self.logger.info("Wrongly introduced {} new reads".format(new_reads_num))

        raw_kept_counts = []
        correct_kept_counts = []
        kept_reads = correct_unique_reads & raw_unique_reads
        for read in kept_reads:
            correct_kept_counts.append(correct_read2count[read])
            raw_kept_counts.append(raw_read2count[read])

        # unique_kept_reads_num = len(kept_reads)
        raw_unique_num = len(raw_unique_reads)
        shared_vars3 = total_num, raw_unique_num
        with WorkerPool(self.config.num_workers, shared_objects=shared_vars3, start_method='fork') as pool:
            raw_kept_entropy_lst = pool.map(self.entropy_item, raw_kept_counts)

        correct_kept_total_num = sum(correct_kept_counts)
        # shared_vars4 = correct_kept_total_num, unique_kept_reads_num
        shared_vars4 = total_num, raw_unique_num
        with WorkerPool(self.config.num_workers, shared_objects=shared_vars4, start_method='fork') as pool:
            correct_kept_entropy_lst = pool.map(self.entropy_item, correct_kept_counts) 

        raw_removed_reads = raw_unique_reads - correct_unique_reads
        raw_removed_items = []
        for read in raw_removed_reads:
            raw_removed_items.append(raw_read2count[read])
        # removed_total_num = sum(raw_removed_items)
        # unique_removed_num = len(raw_removed_reads)
        shared_vars5 = total_num, raw_unique_num
        with WorkerPool(self.config.num_workers, shared_objects=shared_vars5, start_method='fork') as pool:
            raw_removed_entropy_items_lst = pool.map(self.entropy_item, raw_removed_items)
        for i in raw_removed_entropy_items_lst:
            if i <=0:
                print('Warning')
        entropy_item_lst = []
        for i, j in zip(raw_kept_entropy_lst, correct_kept_entropy_lst):
            entropy_item_lst.append(i - j)
        entropy_item_lst.extend(raw_removed_entropy_items_lst)
        if new_reads_num > 0:
            entropy_item_lst.extend([np.nan] * new_reads_num)
        shuffle(entropy_item_lst)
        self.gain2heatmap(entropy_item_lst)
        return [raw_entropy, correct_entropy]

    '''